from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ..lang import build_instance
from ..lang import sig
from ..lang import H
from ..lang import t
from ..Control.Applicative import Applicative
from ..Control.Monad import Monad
from .Foldable import Foldable
from .Functor import Functor


class Traversable(Foldable, Functor):
    """
    Functors representing data structures that can be traversed from left to
    right.

    Dependencies:

    - `~hask.Data.Foldable.Foldable`:class:
    - `~hask.Data.Functor.Functor`:class:

    Attributes:

    - ``traverse``
    - ``sequenceA``
    - ``mapM``
    - ``sequence``


    Minimal complete definition:

    - ``traverse``

    """
    @classmethod
    def make_instance(typeclass, cls, traverse, sequenceA=None, mapM=None,
                      sequence=None):
        attrs = {"traverse": traverse, "sequenceA": sequenceA, "mapM": mapM,
                 "sequence": sequence}
        build_instance(Traversable, cls, attrs)


@sig(H[(Applicative, "f"), (Traversable, "t")]/
        (H/ "a" >> t("f", "b")) >> t("t", "a") >> t("f", t("t", "b")))
def traverse(f, t):
    """``traverse :: (Traversable t, Applicative f) => (a -> f b) -> t a -> f (t b)``

    Map each element of a structure to an action, evaluate these these actions
    from left to right, and collect the results. actions from left to right,
    and collect the results. For a version that ignores the results see
    ``traverse_``.

    """
    return Traversable[t].traverse(f, t)


@sig(H[(Applicative, "f"), (Traversable, "t")]/
        t("t", t("f", "a")) >> t("f", t("t", "a")))
def sequenceA(t):
    """``sequenceA :: (Traversable t, Applicative f) => t (f a) -> f (t a)``

    Evaluate each action in the structure from left to right, and and collect
    the results. For a version that ignores the results see ``sequenceA_``.

    """
    return Traversable[t].sequenceA(t)


@sig(H[(Monad, "m"), (Traversable, "t")]/
        (H/ "a" >> t("m", "b")) >> t("t", "a") >> t("m", t("t", "b")))
def mapM(f, m):
    """``mapM :: (Traversable t, Monad m) => (a -> m b) -> t a -> m (t b)``

    Map each element of a structure to a monadic action, evaluate these
    actions from left to right, and collect the results. For a version that
    ignores the results see ``mapM_``.

    """
    return Traversable[t].mapM(f, t)


@sig(H[(Monad, "m"), (Traversable, "t")]/
        t("t", t("m", "a")) >> t("m", t("t", "a")))
def sequence(t):
    """``sequence :: (Traversable t, Monad m) => t (m a) -> m (t a)``

    Evaluate each monadic action in the structure from left to right, and
    collect the results. For a version that ignores the results see
    ``sequence_``.

    """
    return Traversable[t].sequence(t)


@sig(H[(Applicative, "f"), (Traversable, "t")]/
        t("t", "a") >> (H/ "a" >> t("f", "b")) >> t("f", t("t", "b")))
def for1(t, f):
    """``for1 :: (Traversable t, Applicative f) => t a -> (a -> f b) -> f (t b)``

    for1 is traverse with its arguments flipped. For a version that ignores the
    results see ``for1_``.

    """
    return traverse(f, t)


@sig(H[(Monad, "m"), (Traversable, "t")]/
        t("t", "a") >> (H/ "a" >> t("m", "b")) >> t("m", t("t", "b")))
def forM(t, f):
    """``forM :: (Traversable t, Monad m) => t a -> (a -> m b) -> m (t b)``

    forM is mapM with its arguments flipped. For a version that ignores the
    results see ``forM_``.

    """
    return mapM(f, t)


@sig(H[(Traversable, "t")]/ (H/ "a" >> "b" >> ("a", "c")) >> "a" >> t("t", "b")
        >> ("a", t("t", "c")))
def mapAccumL(f, a, tb):
    """``mapAccumL :: Traversable t => (a -> b -> (a, c)) -> a -> t b -> (a, t c)``

    The mapAccumL function behaves like a combination of fmap and foldl; it
    applies a function to each element of a structure, passing an accumulating
    parameter from left to right, and returning a final value of this
    accumulator together with the new structure.
    """
    raise NotImplementedError()


@sig(H[(Traversable, "t")]/ (H/ "a" >> "b" >> ("a", "c")) >> "a" >> t("t", "b")
        >> ("a", t("t", "c")))
def mapAccumR(f, a, tb):
    """``mapAccumR :: Traversable t => (a -> b -> (a, c)) -> a -> t b -> (a, t c)``

    The mapAccumR function behaves like a combination of fmap and foldr; it
    applies a function to each element of a structure, passing an accumulating
    parameter from right to left, and returning a final value of this
    accumulator together with the new structure.
    """
    raise NotImplementedError()



# flake8: noqa
from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from . import lang
from . import Data
from . import Control
from . import Python


# Core language

## Typeclass instance declaration
from hask.lang import instance

## Operator sections
from hask.lang import __

## Guard expressions
from hask.lang import guard
from hask.lang import c
from hask.lang import otherwise
from hask.lang import NoGuardMatchException

## Lists/list comprehensions
from hask.lang import L

## ADT creation
from hask.lang import data
from hask.lang import d
from hask.lang import deriving

## Type signatures
from hask.lang import sig
from hask.lang import H
from hask.lang import t
from hask.lang import func
from hask.lang import TypeSignatureError

## Pattern matching
from hask.lang import caseof
from hask.lang import p
from hask.lang import m
from hask.lang import IncompletePatternError

## REPL tools
from hask.lang import _t
from hask.lang import _i
from hask.lang import _q

## Type system/typeclasses
from hask.lang import typeof
from hask.lang import has_instance
from hask.lang import Typeclass
from hask.lang import Hask


# Basic Typeclasses
from hask.Prelude import Read
from hask.Prelude import Show
from hask.Prelude import Eq
from hask.Prelude import Ord
from hask.Prelude import Enum
from hask.Prelude import Bounded
from hask.Prelude import Num
from hask.Prelude import Real
from hask.Prelude import Integral
from hask.Prelude import Fractional
from hask.Prelude import Floating
from hask.Prelude import RealFrac
from hask.Prelude import RealFloat
from hask.Prelude import Functor
from hask.Prelude import Applicative
from hask.Prelude import Monad
from hask.Prelude import Traversable
from hask.Prelude import Foldable


# Standard types
from hask.Prelude import Maybe
from hask.Prelude import Just
from hask.Prelude import Nothing
from hask.Prelude import in_maybe

from hask.Prelude import Either
from hask.Prelude import Left
from hask.Prelude import Right
from hask.Prelude import in_either

from hask.Prelude import Ordering
from hask.Prelude import LT
from hask.Prelude import EQ
from hask.Prelude import GT



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ..lang import H
from ..lang import sig


@sig(H/ ("a", "b") >> "a")
def fst(tup):
    """``fst :: (a, b) -> a``

    Extract the first component of a pair.

    """
    x, y = tup
    return x


@sig(H/ ("a", "b") >> "b")
def snd(tup):
    """``snd :: (a, b) -> b``

    Extract the second component of a pair.

    """
    x, y = tup
    return y


@sig(H/ (H/ ("a", "b") >> "c") >> "a" >> "b" >> "c")
def curry(tup_fn, x, y):
    """``curry :: ((a, b) -> c) -> a -> b -> c``

    curry converts an uncurried function to a curried function.

    """
    return tup_fn((x, y))


@sig(H/ (H/ "a" >> "b" >> "c") >> ("a", "b") >> "c")
def uncurry(fn, tup):
    """``uncurry :: (a -> b -> c) -> (a, b) -> c``

    uncurry converts a curried function to a function on pairs.

    """
    return fn(fst(tup), snd(tup))


@sig(H/ ("a", "b") >> ("b", "a"))
def swap(tup):
    """``swap :: (a, b) -> (b, a)``

    Swap the components of a pair.

    """
    a, b = tup
    return (b, a)



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ..lang import Show
from ..lang import Read
from ..lang import Bounded
from ..lang import Ord
from ..lang import sig
from ..lang import H
from ..lang import data
from ..lang import d
from ..lang import deriving

from .Eq import Eq


# data Ordering = LT | EQ | GT deriving(Show, Eq, Ord, Bounded)
Ordering, LT, EQ, GT =\
        data.Ordering == d.LT | d.EQ | d.GT & deriving(Read, Show, Eq, Ord, Bounded)


# TODO: Down?


@sig(H[(Ord, "a")]/ "a" >> "a" >> "a")
def max(x, y):
    """``max :: a -> a -> a``

    Maximum function.

    """
    return x if x >= y else y


@sig(H[(Ord, "a")]/ "a" >> "a" >> "a")
def min(x, y):
    """``min :: a -> a -> a``

    Minumum function.

    """
    return x if x <= y else y


@sig(H[(Ord, "a")]/ "a" >> "a" >> Ordering)
def compare(x, y):
    """``compare :: a -> a -> Ordering``

    Comparison function.

    """
    return EQ if x == y else (LT if x < y else GT)


@sig(H[(Ord, "a")]/ (H/ "a" >> "b") >> "b" >> "b" >> Ordering)
def comparing(p, x, y):
    """``comparing :: Ord a => (b -> a) -> b -> b -> Ordering``

    comparing(p, x, y) = compare(p(x), p(y))

    Useful combinator for use in conjunction with the `xxxBy` family of
    functions from Data.List, for example::

       ... sortBy (comparing(fst)) ...

    """
    return compare(p(x), p(y))



# flake8: noqa
from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import fractions

from .lang import H
from .lang import sig
from .lang import t

from .Data.Maybe import Maybe
from .Data.Maybe import Just
from .Data.Maybe import Nothing
from .Data.Maybe import in_maybe
from .Data.Maybe import maybe

from .Data.Either import Either
from .Data.Either import Left
from .Data.Either import Right
from .Data.Either import in_either
from .Data.Either import either

from .Data.Ord import Ordering
from .Data.Ord import LT
from .Data.Ord import EQ
from .Data.Ord import GT

from .Data.Tuple import fst
from .Data.Tuple import snd
from .Data.Tuple import curry
from .Data.Tuple import uncurry


from .lang import Read
from .lang import Show
from .lang import show

from .Data.Eq import Eq
from .Data.Ord import Ord
from .Data.Ord import max
from .Data.Ord import min
from .Data.Ord import compare

from .lang import Enum
from .lang import fromEnum
from .lang import succ
from .lang import pred
from .lang import enumFromThen
from .lang import enumFrom
from .lang import enumFromThenTo
from .lang import enumFromTo

from .lang import Bounded
from .Data.Functor import Functor
from .Data.Functor import fmap

from .Control.Applicative import Applicative
from .Control.Monad import Monad
from .Data.Foldable import Foldable
from .Data.Traversable import Traversable

from .Data.Num import Num
from .Data.Num import abs
from .Data.Num import negate
from .Data.Num import signum

from .Data.Num import Fractional
from .Data.Num import recip

from .Data.Num import Integral
from .Data.Num import toRatio

from .Data.Num import Ratio
from .Data.Num import R
from .Data.Num import Rational

from .Data.Num import Floating
from .Data.Num import exp
from .Data.Num import sqrt
from .Data.Num import log
from .Data.Num import pow
from .Data.Num import logBase
from .Data.Num import sin
from .Data.Num import tan
from .Data.Num import cos
from .Data.Num import asin
from .Data.Num import atan
from .Data.Num import acos
from .Data.Num import sinh
from .Data.Num import tanh
from .Data.Num import cosh
from .Data.Num import asinh
from .Data.Num import atanh
from .Data.Num import acosh

from .Data.Num import Real
from .Data.Num import toRational

from .Data.Num import RealFrac
from .Data.Num import properFraction
from .Data.Num import truncate
from .Data.Num import round
from .Data.Num import ceiling
from .Data.Num import floor

from .Data.Num import RealFloat
from .Data.Num import isNaN
from .Data.Num import isInfinite
from .Data.Num import isNegativeZero
from .Data.Num import atan2


@sig(H[(Num, "a")]/ "a" >> "a" >> "a")
def subtract(x, y):
    """
    subtract :: Num a => a -> a -> a

    the same as lambda x, y: y - x
    """
    return y - x


@sig(H[(Integral, "a")]/ "a" >> bool)
def even(x):
    """
    even :: Integral a => a -> Bool

    Returns True if the integral value is even, and False otherwise.
    """
    return x % 2 == 0


@sig(H[(Integral, "a")]/ "a" >> bool)
def odd(x):
    """
    odd :: Integral a => a -> Bool

    Returns True if the integral value is odd, and False otherwise.
    """
    return x % 2 == 1


@sig(H[(Integral, "a")]/ "a" >> "a" >> "a")
def gcd(x, y):
    """
    gcd :: Integral a => a -> a -> a

    gcd(x,y) is the non-negative factor of both x and y of which every common
    factor of x and y is also a factor; for example gcd(4,2) = 2, gcd(-4,6) =
    2, gcd(0,4) = 4. gcd(0,0) = 0. (That is, the common divisor that is
    "greatest" in the divisibility preordering.)
    """
    return fractions.gcd(x, y)


@sig(H[(Integral, "a")]/ "a" >> "a" >> "a")
def lcm(x, y):
    """
    lcm :: Integral a => a -> a -> a

    lcm(x,y) is the smallest positive integer that both x and y divide.
    """
    g = gcd(x, y)
    return 0 if g == 0 else (x * y) // g


from .Data.Functor import Functor
from .Control.Applicative import Applicative
from .Control.Monad import Monad


@sig(H[(Monad, "m")]/ t("m", "a") >> t("m", None))
def sequence(xs):
    """
    sequence :: Monad m => [m a] -> m [a]

    Evaluate each action in the sequence from left to right, and collect the
    results.
    """
    raise NotImplementedError()


@sig(H[(Monad, "m")]/ t("m", "a") >> t("m", None))
def sequence_(xs):
    """
    sequence_ :: Monad m => [m a] -> m None

    Evaluate each action in the sequence from left to right, and ignore the
    results.
    """
    raise NotImplementedError()


def mapM(f, xs):
    """
    mapM :: Monad m => (a -> m b) -> [a] -> m [b]

    mapM(f) is equivalent to sequence * map(f)
    """
    return sequence(fmap(f, xs))


def mapM_(f, xs):
    """
    mapM_ :: Monad m => (a -> m b) -> [a] -> m ()

    mapM_(f) is equivalent to sequence_ * map(f)
    """
    return sequence_(fmap(f, xs))


@sig(H/ "a" >> "a")
def id(a):
    """
    id :: a -> a

    Identity function.
    """
    return a


@sig(H/ "a" >> "b" >> "a")
def const(a, b):
    """
    const :: a -> b -> a

    Constant function.
    """
    return a


@sig(H/ (H/ "a" >> "b" >> "c") >> "b" >> "a" >> "c")
def flip(f, b, a):
    """
    flip :: (a -> b -> c) -> b -> a -> c

    flip(f) takes its (first) two arguments in the reverse order of f.
    """
    return f(a, b)


@sig(H/ (H/ "a" >> bool) >> (H/ "a" >> "a") >> "a" >> "a")
def until(p, f, a):
    """
    until :: (a -> Bool) -> (a -> a) -> a -> a

    until(p, f, a) yields the result of applying f until p(a) holds.
    """
    while not p(a):
        a = f(a)
    return a


@sig(H/ "a" >> "a" >> "a")
def asTypeOf(a, b):
    """
    asTypeOf :: a -> a -> a

    asTypeOf is a type-restricted version of const. It is usually used as an
    infix operator, and its typing forces its first argument (which is usually
    overloaded) to have the same type as the second.
    """
    return a


@sig(H/ str >> "a")
def error(msg):
    """
    error :: str -> a

    error(msg) stops execution and displays an error message.
    """
    raise Exception(msg)


from .lang import undefined


from .Data.List import map
from .Data.List import filter
from .Data.List import head
from .Data.List import last
from .Data.List import tail
from .Data.List import init
from .Data.List import null
from .Data.List import reverse
from .Data.List import length

from .Data.List import foldl
from .Data.List import foldl1
from .Data.List import foldr
from .Data.List import foldr1


from .Data.List import and_
from .Data.List import or_
from .Data.List import any
from .Data.List import all
from .Data.List import sum
from .Data.List import product
from .Data.List import concat
from .Data.List import concatMap
from .Data.List import maximum
from .Data.List import minimum


from .Data.List import scanl
from .Data.List import scanl1
from .Data.List import scanr
from .Data.List import scanr1


from .Data.List import iterate
from .Data.List import repeat
from .Data.List import replicate
from .Data.List import cycle


from .Data.List import take
from .Data.List import drop
from .Data.List import splitAt
from .Data.List import takeWhile
from .Data.List import dropWhile
from .Data.List import span
from .Data.List import break_


from .Data.List import elem
from .Data.List import notElem
from .Data.List import lookup


from .Data.List import zip
from .Data.List import zip3
from .Data.List import zipWith
from .Data.List import zipWith3
from .Data.List import unzip
from .Data.List import unzip3

from .Data.List import lines
from .Data.List import words
from .Data.List import unlines
from .Data.List import unwords



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ..lang import sig
from ..lang import H
from ..lang import t
from ..lang import L
from ..lang import Typeclass
from ..lang import build_instance
from ..lang import is_builtin
from ..lang import List
from ..lang import instance

from . import List as DL
from ..Control.Applicative import Applicative
from ..Control.Monad import Monad
from .Eq import Eq
from .Num import Num
from .Maybe import Maybe
from .Ord import Ord
from .Ord import Ordering


class Foldable(Typeclass):
    """Data structures that can be folded.

    Attributes:

    - ``foldr``
    - ``foldr1``
    - ``foldl``
    - ``foldl_``
    - ``foldl1``
    - ``toList``
    - ``null``
    - ``length``
    - ``elem``
    - ``maximum``
    - ``minimum``
    - ``sum``
    - ``product``

    Minimal complete definition:

    - ``foldr``

    Magic methods:

    - ``__iter__``

    """
    @classmethod
    def make_instance(typeclass, cls, foldr, foldr1=None, foldl=None,
            foldl_=None, foldl1=None, toList=None, null=None, length=None,
            elem=None, maximum=None, minimum=None, sum=None, product=None):

        # attributes that are not supplied are implemented in terms of toList
        if toList is None:
            if hasattr(cls, "__iter__"):
                toList = lambda x: L[iter(x)]
            else:
                toList = lambda t: foldr(lambda x, y: x ^ y, L[[]], t)

        foldr1 = (lambda x: DL.foldr1(toList(x))) if foldr1 is None else foldr1
        foldl = (lambda x: DL.foldl(toList(x))) if foldl is None else foldl
        foldl_ = (lambda x: DL.foldl_(toList(x))) if foldl_ is None else foldl_
        foldl1 = (lambda x: DL.foldl1(toList(x))) if foldl1 is None else foldl1
        null = (lambda x: DL.null(toList(x))) if null is None else null
        length = (lambda x: DL.length(toList(x))) if length is None else length
        elem = (lambda x: DL.length(toList(x))) if length is None else length
        mi = (lambda x: DL.minimum(toList(x))) if minimum is None else minimum
        ma = (lambda x: DL.maximum(toList(x))) if maximum is None else maximum
        sum = (lambda x: DL.sum(toList(x))) if sum is None else sum
        p = (lambda x: DL.product(toList(x))) if product is None else product

        attrs = {"foldr": foldr, "foldr1": foldr1, "foldl": foldl,
                 "foldl_": foldl_, "foldl1": foldl1, "toList": toList,
                 "null": null, "length": length, "elem": elem, "maximum": ma,
                 "minimum": mi, "sum": sum, "product": p}
        build_instance(Foldable, cls, attrs)

        if not hasattr(cls, "__len__") and not is_builtin(cls):
            cls.__len__ = length

        if not hasattr(cls, "__iter__") and not is_builtin(cls):
            cls.__iter__ = lambda x: iter(toList(x))


@sig(H[(Foldable, "t")]/ (H/ "a" >> "b" >> "b") >> "b" >> t("t", "a") >> "b")
def foldr(f, z, t):
    """``foldr :: Foldable t => (a -> b -> b) -> b -> t a -> b``

    Right-associative fold of a structure.

    """
    return Foldable[t].foldr(f, z, t)


@sig(H[(Foldable, "t")]/ (H/ "a" >> "a" >> "a") >> t("t", "a") >> "a")
def foldr1(f, t):
    """``foldr1 :: Foldable t => (a -> a -> a) -> t a -> a``

    A variant of foldr that has no base case, and thus may only be applied to
    non-empty structures.

    """
    return Foldable[t].foldr(f, t)


@sig(H[(Foldable, "t")]/ (H/ "a" >> "a" >> "b") >> "b" >> t("t", "a") >> "b")
def foldl(f, z, t):
    """``foldl :: Foldable t => (b -> a -> b) -> b -> t a -> b``

    Left-associative fold of a structure.

    """
    return Foldable[t].foldl(f, z, t)


@sig(H[(Foldable, "t")]/ (H/ "a" >> "a" >> "b") >> "b" >> t("t", "a") >> "b")
def foldl_(f, z, t):
    """``foldl_ :: Foldable t => (b -> a -> b) -> b -> t a -> b``

    Left-associative fold of a structure. but with strict application of the
    operator.

    """
    return Foldable[t].foldl_(f, z, t)


@sig(H[(Foldable, "t")]/ (H/ "a" >> "a" >> "a") >> t("t", "a") >> "a")
def foldl1(f, t):
    """``foldl1 :: Foldable t => (a -> a -> a) -> t a -> a``

    A variant of foldl that has no base case, and thus may only be applied to
    non-empty structures.

    """
    Foldable[t].foldl1(f, t)


@sig(H[(Foldable, "t")]/ t("t", "a") >> ["a"])
def toList(t):
    """``toList :: Foldable t => t a -> [a]``

    List of elements of a structure, from left to right.

    """
    return Foldable[t].toList(t)


@sig(H[(Foldable, "t")]/ t("t", "a") >> bool)
def null(t):
    """``null :: Foldable t => t a -> Bool``

    Test whether the structure is empty.

    """
    return Foldable[t].null(t)


@sig(H[(Foldable, "t")]/ t("t", "a") >> int)
def length(t):
    """``length :: Foldable t => t a -> int``

    Returns the size/length of a finite structure as an int.

    """
    return Foldable[t].length(t)


@sig(H[(Foldable, "t"), (Eq, "a")]/ t("t", "a") >> "a")
def elem(a, t):
    """``elem :: (Foldable t, Eq a) => a -> t a -> bool``

    Does the element occur in the structure?

    """
    return Foldable[t].elem(t)


@sig(H[(Foldable, "t"), (Ord, "a")]/ t("t", "a") >> "a")
def maximum(t):
    """``maximum :: (Foldable t, forall a. Ord a) => t a -> a``

    The largest element of a non-empty structure.

    """
    return Foldable[t].maximum(t)


@sig(H[(Foldable, "t"), (Ord, "a")]/ t("t", "a") >> "a")
def minimum(t):
    """``minimum :: (Foldable t, forall a. Ord a) => t a -> a``

    The least element of a non-empty structure.

    """
    return Foldable[t].minimum(t)


@sig(H[(Foldable, "t"), (Num, "a")]/ t("t", "a") >> "a")
def sum(t):
    """``sum :: (Foldable t, Num a) => t a -> a``

    The sum function computes the sum of the numbers of a structure.

    """
    return Foldable[t].sum(t)


@sig(H[(Foldable, "t"), (Num, "a")]/ t("t", "a") >> "a")
def product(t):
    """``product :: (Foldablet, Num a) => t a -> a``

    The product function computes the product of the numbers of a structure.

    """
    return Foldable[t].product(t)


@sig(H[(Foldable, "t"), (Monad, "m")]/ (H/ "a" >> "b" >> t("m", "b")) >>
    "b" >> t("t", "a") >> t("m", "b"))
def foldlM(f, b, ta):
    """``foldrM :: (Foldable t, Monad m) => (a -> b -> m b) -> b -> t a -> m b``

    Monadic fold over the elements of a structure, associating to the right,
    i.e. from right to left.

    """
    raise NotImplementedError()


@sig(H[(Foldable, "t"), (Monad, "m")]/ (H/ "b" >> "a" >> t("m", "b")) >>
    "b" >> t("t", "a") >> t("m", "b"))
def foldrM(f, b, ta):
    """``foldlM :: (Foldable t, Monad m) => (b -> a -> m b) -> b -> t a -> m b``

    Monadic fold over the elements of a structure, associating to the left,
    i.e. from left to right.

    """
    raise NotImplementedError()


@sig(H[(Foldable, "t"), (Applicative, "f")]/ (H/ "a" >> t("f", "b")) >>
        t("f", "a") >> t("f", None))
def traverse_(f, t):
    """``traverse_ :: (Foldable t, Applicative f) => (a -> f b) -> t a -> f ()``

    Map each element of a structure to an action, evaluate these actions from
    left to right, and ignore the results.  For a version that doesn't ignore
    the results see traverse.

    """
    raise NotImplementedError()


@sig(H[(Foldable, "t"), (Applicative, "f")]/ t("f", "a") >>
        (H/ "a" >> t("f", "b")) >> t("f", None))
def for_(t, f):
    """``for_ :: (Foldable t, Applicative f) => t a -> (a -> f b) -> f ()``

    ``for_`` is ``traverse_`` with its arguments flipped.  For a version that
    doesn't ignore the results see for.

    """
    return traverse_(f, t)


@sig(H[(Foldable, "t"), (Applicative, "f")]/ t("t", t("m", "a")) >>
        t("f", None))
def sequenceA_(t):
    """``sequenceA_ :: (Foldable t, Applicative f) => t (f a) -> f ()``

    Evaluate each action in the structure from left to right, and ignore the
    results.  For a version that doesn't ignore the results see sequenceA.

    """
    raise NotImplementedError()


@sig(H[(Foldable, "t"), (Monad, "m")]/ (H/ "a" >> t("m", "b")) >>
        t("m", "a") >> t("m", None))
def mapM_(f, t):
    """``mapM_ :: (Foldable t, Monad m) => (a -> m b) -> t a -> m ()``

    Map each element of a structure to a monadic action, evaluate these
    actions from left to right, and ignore the results.  For a version that
    doesn't ignore the results see mapM.

    As of base 4.8.0.0, ``mapM_`` is just ``traverse_``, specialized to Monad.

    """
    return traverse_(f, t)


@sig(H[(Foldable, "t"), (Monad, "m")]/ t("m", "a") >>
        (H/ "a" >> t("m", "b")) >> t("m", None))
def forM_(t, f):
    """``forM_ :: (Foldable t, Monad m) => t a -> (a -> m b) -> m ()``

    ``forM_`` is ``mapM_`` with its arguments flipped.  For a version that
    doesn't ignore the results see forM.

    As of base 4.8.0.0, ``forM_`` is just ``for_``, specialized to Monad.

    """
    return mapM_(f, t)


@sig(H[(Foldable, "t"), (Monad, "m")]/ t("t", t("m", "a")) >> t("m", None))
def sequence_(t):
    """``sequence_ :: (Foldable t, Monad m) => t (m a) -> m ()``

    Evaluate each monadic action in the structure from left to right, and
    ignore the results.  For a version that doesn't ignore the results see
    sequence.

    As of base 4.8.0.0, ``sequence_`` is just ``sequenceA_``, specialized to
    Monad.

    """
    return sequenceA_(t)


@sig(H[(Foldable, "t")]/ t("t", ["a"]) >> ["a"])
def concat(t):
    """``concat :: Foldable t => t [a] -> [a]``

    The concatenation of all the elements of a container of lists.

    """
    return DL.concat(toList(t))


@sig(H[(Foldable, "t")]/ (H/ "a" >> ["b"]) >> t("t", "a") >> ["b"])
def concatMap(f, t):
    """``concatMap :: Foldable t => (a -> [b]) -> t a -> [b]``

    Map a function over all the elements of a container and concatenate the
    resulting lists.

    """
    return DL.concatMap(f, toList(t))


@sig(H[(Foldable, "t")]/ t("t", bool) >> bool)
def and_(t):
    """``and :: Foldable t => t bool -> bool``

    and returns the conjunction of a container of Bools.  For the result to be
    True, the container must be finite; False, however, results from a False
    value finitely far from the left end.

    """
    return DL.and_(toList(t))


@sig(H[(Foldable, "t")]/ t("t", bool) >> bool)
def or_(t):
    """``or :: Foldable t => t bool -> bool``

    or returns the disjunction of a container of Bools.  For the result to be
    False, the container must be finite; True, however, results from a True
    value finitely far from the left end.

    """
    return DL.or_(toList(t))


@sig(H[(Foldable, "t")]/ (H/ "a" >> bool) >> t("t", "a") >> bool)
def any_(f, t):
    """``any :: Foldable t => (a -> bool) -> t a -> bool``

    Determines whether any element of the structure satisfies the predicate.

    """
    return DL.any_(toList(t))


@sig(H[(Foldable, "t")]/ (H/ "a" >> bool) >> t("t", "a") >> bool)
def all_(f, t):
    """``all :: Foldable t => (a -> bool) -> t a -> bool``

    Determines whether all elements of the structure satisfy the predicate.

    """
    return DL.all_(toList(t))


@sig(H[(Foldable, "t")]/ (H/ "a" >> "a" >> Ordering) >> t("t", "a") >> "a")
def maximumBy_(f, t):
    """``maximumBy :: Foldable t => (a -> a -> Ordering) -> t a -> a``

    The largest element of a non-empty structure with respect to the given
    comparison function.

    """
    return DL.maximumBy(toList(t))


@sig(H[(Foldable, "t")]/ (H/ "a" >> "a" >> Ordering) >> t("t", "a") >> "a")
def minimumBy_(f, t):
    """``minimumBy :: Foldable t => (a -> a -> Ordering) -> t a -> a``

    The least element of a non-empty structure with respect to the given
    comparison function.

    """
    return DL.minimumBy(toList(t))


@sig(H[(Foldable, "t"), (Eq, "a")]/ "a" >> t("t", "a") >> bool)
def notElem(a, t):
    """``notElem :: (Foldable t, Eq a) => a -> t a -> bool``

    notElem is the negation of elem.

    """
    return not elem(a, t)


@sig(H[(Foldable, "t")]/ (H/ "a" >> bool) >> t("t", "a") >> t(Maybe, "a"))
def find(f, t):
    """``find :: Foldable t => (a -> bool) -> t a -> Maybe a``

    The find function takes a predicate and a structure and returns the
    leftmost element of the structure matching the predicate, or Nothing if
    there is no such element.

    """
    return DL.find(f, toList(t))


instance(Foldable, List).where(
    foldr = DL.foldr,
    foldr1 = DL.foldr1,
    foldl = DL.foldl,
    foldl_ = DL.foldl_,
    foldl1 = DL.foldl1,
    null = DL.null,
    length = DL.length,
    elem = DL.elem,
    minimum = DL.minimum,
    maximum = DL.maximum,
    sum = DL.sum,
    product = DL.product
)



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import itertools

from ..lang import sig
from ..lang import H
from ..lang import t
from ..lang import L
from ..lang import build_instance
from ..lang import is_builtin
from ..lang import List
from ..lang import instance
from ..Data.Functor import fmap
from .Applicative import Applicative


class Monad(Applicative):
    """The Monad class defines the basic operations over a monad, a concept from
    a branch of mathematics known as category theory.  From the perspective of
    a Haskell programmer, however, it is best to think of a monad as an
    abstract datatype of actions.

    Dependencies:

    - `~hask.Control.Applicative.Applicative`:class:

    Attributes:

    - ``bind``
    - ``__rshift__``

    Minimal complete definition:

    - ``bind``


    """
    @classmethod
    def make_instance(typeclass, cls, bind):
        bind = bind ** (H[(Monad, "m")]/ t("m", "a") >> (H/ "a" >> t("m", "b"))
                >> t("m", "b"))
        if not is_builtin(cls):
            def bind_wrap(s, o):
                return Monad[s].bind(s, o)
            cls.__rshift__ = bind_wrap
        build_instance(Monad, cls, {"bind": bind})


@sig(H[(Monad, "m")]/ t("m", "a") >> (H/ "a" >> t("m", "b")) >> t("m", "b"))
def bind(m, fn):
    """
    ``bind :: Monad m => m a -> (a -> m b) -> m b``

    Monadic bind.
    """
    return Monad[m].bind(m, fn)


@sig(H[(Monad, "m")]/ t("m", t("m", "a")) >> t("m", "a"))
def join(m):
    """``join :: Monad m => m (m a) -> m a``

    The join function is the conventional monad join operator.  It is used to
    remove one level of monadic structure, projecting its bound argument into
    the outer level.

    """
    __id = (lambda x: x) ** (H/ "a" >> "a")
    return bind(m, __id)


@sig(H[(Monad, "m")]/ (H/ "a" >> "r") >> t("m", "a") >> t("m", "r"))
def liftM(fn, m):
    """
    ``liftM :: Monad m => (a1 -> r) -> m a1 -> m r``

    Promote a function to a monad.

    """
    return fmap(fn, m)


instance(Monad, List).where(
    bind = lambda x, fn: L[itertools.chain.from_iterable(fmap(fn, x))]
)



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ..lang import build_instance
from ..lang import List, L
from ..lang import instance
from ..Data.Functor import Functor


class Applicative(Functor):
    """
    A functor with application, providing operations to embed pure expressions
    (pure), and sequence computations and combine their results (ap).

    Dependencies:

    - `~hask.Data.Functor.Functor`:class:

    Attributes:

    - ``pure``

    Minimal complete definition:

    - ``pure``

    """
    @classmethod
    def make_instance(self, cls, pure):
        build_instance(Applicative, cls, {"pure": pure})


instance(Applicative, List).where(
    pure = lambda x: L[[x]]
)



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import math
import fractions
import sys

from ..lang import data
from ..lang import d
from ..lang import deriving
from ..lang import H
from ..lang import sig
from ..lang import t
from ..lang import instance
from ..lang import build_instance
from ..lang import Enum
from ..lang import Show
from .Eq import Eq
from .Ord import Ord


class Num(Show, Eq):
    """Basic numeric class.

    Dependencies:

    - `~hask.lang.typeclasses.Show`:class:
    - `~hask.lang.typeclasses.Eq`:class:

    Attributes:

    - ``__add__``
    - ``__mul__``
    - ``__abs__``
    - ``signum``
    - ``fromInteger``
    - ``__neg__``
    - ``__sub__``

    Minimal complete definition:

    - ``add``
    - ``mul``
    - ``abs``
    - ``signum``
    - ``fromInteger``
    - ``negate``

    """
    @classmethod
    def make_instance(typeclass, cls, add, mul, abs, signum, fromInteger,
                      negate, sub=None):

        @sig(H[(Num, "a")]/ "a" >> "a" >> "a")
        def default_sub(a, b):
            return a.__add__(b.__neg__())

        sub = default_sub if sub is None else sub
        attrs = {"add": add, "mul": mul, "abs": abs, "signum": signum,
                 "fromInteger": fromInteger, "negate": negate, "sub": sub}

        build_instance(Num, cls, attrs)


@sig(H[(Num, "a")]/ "a" >> "a")
def negate(a):
    """``signum :: Num a => a -> a``

    Unary negation.

    """
    return Num[a].negate(a)


@sig(H[(Num, "a")]/ "a" >> "a")
def signum(a):
    """``signum :: Num a => a -> a``

    Sign of a number. The functions abs and signum should satisfy the law:
    abs x * signum x == x
    For real numbers, the signum is either -1 (negative), 0 (zero) or 1
    (positive).

    """
    return Num[a].signum(a)


@sig(H[(Num, "a")]/ "a" >> "a")
def abs(a):
    """``abs :: Num a => a -> a``

    Absolute value.

    """
    return Num[a].abs(a)


instance(Num, int).where(
    add = int.__add__,
    mul = int.__mul__,
    abs = int.__abs__,
    signum = lambda x: -1 if x < 0 else (1 if x > 0 else 0),
    fromInteger = int,
    negate = int.__neg__,
    sub = int.__sub__
)

try:
    instance(Num, long).where(
        add = long.__add__,
        mul = long.__mul__,
        abs = long.__abs__,
        signum = lambda x: long(-1) if x < long(0) else (long(1) if x > long(0) else long(0)),
        fromInteger = long,
        negate = long.__neg__,
        sub = long.__sub__
    )
except NameError:
    pass

instance(Num, float).where(
    add = float.__add__,
    mul = float.__mul__,
    abs = float.__abs__,
    signum = lambda x: -1.0 if x < 0.0 else (1.0 if x > 0.0 else 0.0),
    fromInteger = float,
    negate = float.__neg__,
    sub = float.__sub__
)

instance(Num, complex).where(
    add = complex.__add__,
    mul = complex.__mul__,
    abs = complex.__abs__,
    signum = lambda x: complex(-1) if x < 0 else complex(1 if x > 0 else 0),
    fromInteger = complex,
    negate = complex.__neg__,
    sub = complex.__sub__
)


class Fractional(Num):
    """Fractional numbers, supporting real division.

    Dependencies:

    - `Num`:class:

    Attributes:

    - ``fromRational``
    - ``recip``
    - ``__div__``

    Minimal complete definition:

    - ``fromRational``
    - ``div``

    """
    @classmethod
    def make_instance(typeclass, cls, fromRational, div, recip=None):
        if recip is None:
            recip = lambda x: div(1, x)
        attrs = {"fromRational": fromRational, "div": div, "recip": recip}
        build_instance(Fractional, cls, attrs)


Ratio, R =\
        data.Ratio("a") == d.R("a", "a") & deriving(Eq)

Rational = t(Ratio, int)


instance(Fractional, float).where(
    fromRational = lambda rat: float(rat[0])/float(rat[1]),
    div = lambda a, b: float(a)/float(b),
    recip = lambda x: 1.0/x
)


@sig(H[(Fractional, "a")]/ "a" >> "a")
def recip(a):
    """``recip :: Fractional a => a -> a``

    Reciprocal fraction.

    """
    return Fractional[a].recip(a)


class Floating(Fractional):
    """Trigonometric and hyperbolic functions and related functions.

    Dependencies:

    - `Fractional`:class:

    Attributes:

    - ``pi``
    - ``exp``
    - ``sqrt``
    - ``log``
    - ``pow``
    - ``logBase``
    - ``sin``
    - ``tan``
    - ``cos``
    - ``asin``
    - ``atan``
    - ``acos``
    - ``sinh``
    - ``tanh``
    - ``cosh``
    - ``asinh``
    - ``atanh``
    - ``acosh``

    Minimal complete definition:

    - ``pi``
    - ``exp``
    - ``sqrt``
    - ``log``
    - ``pow``
    - ``logBase``
    - ``sin``
    - ``tan``
    - ``cos``
    - ``asin``
    - ``atan``
    - ``acos``
    - ``sinh``
    - ``tanh``
    - ``cosh``
    - ``asinh``
    - ``atanh``
    - ``acosh``

    """
    @classmethod
    def make_instance(typeclass, cls, pi, exp, sqrt, log, pow, logBase, sin,
            tan, cos, asin, atan, acos, sinh, tanh, cosh, asinh, atanh, acosh):
        attrs = {"pi": pi, "exp": exp, "sqrt": sqrt, "log": log, "pow": pow,
                "logBase": logBase, "sin": sin, "tan": tan, "cos": cos,
                "asin": asin, "atan": atan, "acos": acos, "sinh": sinh,
                "tanh": tanh, "cosh": cosh, "asinh": asinh, "atanh": atanh,
                "acosh": acosh}
        build_instance(Floating, cls, attrs)


@sig(H[(Floating, "a")]/ "a" >> "a")
def exp(x):
    """``exp :: Floating a => a -> a``

    """
    return Floating[x].exp(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def sqrt(x):
    """``sqrt :: Floating a => a -> a``

    """
    return Floating[x].sqrt(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def log(x):
    """``log:: Floating a => a -> a``

    """
    return Floating[x].log(x)


@sig(H[(Floating, "a")]/ "a" >> "a" >> "a")
def pow(x, y):
    """``pow :: Floating a => a -> a -> a``

    """
    return Floating[x].pow(x, y)


@sig(H[(Floating, "a")]/ "a" >> "a" >> "a")
def logBase(x, b):
    """
    ``logBase :: Floating a => a -> a -> a``
    """
    return Floating[x].logBase(x, b)


@sig(H[(Floating, "a")]/ "a" >> "a")
def sin(x):
    """
    ``sin :: Floating a => a -> a``
    """
    return Floating[x].sin(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def cos(x):
    """
    ``cos :: Floating a => a -> a``
    """
    return Floating[x].cos(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def tan(x):
    """
    ``tan :: Floating a => a -> a``
    """
    return Floating[x].tan(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def asin(x):
    """
    ``asin :: Floating a => a -> a``
    """
    return Floating[x].asin(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def atan(x):
    """
    ``atan :: Floating a => a -> a``
    """
    return Floating[x].atan(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def acos(x):
    """
    ``acos :: Floating a => a -> a``
    """
    return Floating[x].acos(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def sinh(x):
    """
    ``sinh :: Floating a => a -> a``
    """
    return Floating[x].sinh(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def tanh(x):
    """
    ``tanh :: Floating a => a -> a``
    """
    return Floating[x].tanh(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def cosh(x):
    """
    ``cosh :: Floating a => a -> a``
    """
    return Floating[x].cosh(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def asinh(x):
    """``asinh :: Floating a => a -> a``

    """
    return Floating[x].asinh(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def atanh(x):
    """
    ``atanh :: Floating a => a -> a``
    """
    return Floating[x].atanh(x)


@sig(H[(Floating, "a")]/ "a" >> "a")
def acosh(x):
    """
    ``acosh :: Floating a => a -> a``
    """
    return Floating[x].acosh(x)


instance(Floating, float).where(
    pi = math.pi,
    exp = math.exp,
    sqrt = math.sqrt,
    log = math.log,
    pow = pow,
    logBase = math.log,
    sin = math.sin,
    tan = math.tan,
    cos = math.cos,
    asin = math.asin,
    atan = math.atan,
    acos = math.acos,
    sinh = math.sinh,
    tanh = math.tanh,
    cosh = math.cosh,
    asinh = math.asinh,
    atanh = math.atanh,
    acosh = math.acosh
)


class Real(Num, Ord):
    """Real numbers.

    Dependencies:

    - `Num`:class:
    - `~hask.lang.typeclasses.Ord`:class:

    Attributes:

    - ``toRational``

    Minimal complete definition:

    - ``toRational``

    """
    @classmethod
    def make_instance(typeclass, cls, toRational):
        build_instance(Real, cls, {})


@sig(H[(Real, "a")]/ "a" >> Rational)
def toRational(x):
    """``toRational :: Real a => a -> Rational``

    Conversion to Rational.

    """
    return Real[x].toRational(x)


class Integral(Real, Enum):
    """Integral numbers, supporting integer division.

    Dependencies:

    - `Real`:class:
    - `~hask.lang.lazylist.Enum`:class:

    Attributes:

    - ``quotRem``
    - ``toInteger``
    - ``quot``
    - ``rem``
    - ``div``
    - ``mod``

    Minimal complete definition:

    - ``quotRem``
    - ``toInteger``
    - ``quot``
    - ``rem``
    - ``div``
    - ``mod``

    """
    @classmethod
    def make_instance(typeclass, cls, quotRem, divMod, toInteger, quot=None,
            rem=None, div=None, mod=None):

        quot = lambda x: quotRem(x)[0] if quot is None else quot
        rem = lambda x: quotRem(x)[1] if rem is None else rem
        div = lambda x: divMod(x)[0] if div is None else div
        mod = lambda x: divMod(x)[1] if mod is None else mod

        attrs = {"quotRem": quotRem, "toInteger": toInteger, "quot": quot,
                 "rem": rem, "div": div, "mod": mod, "divMod": divMod}
        build_instance(Integral, cls, attrs)


@sig(H[(Integral, "a")]/ "a" >> "a" >> t(Ratio, "a"))
def toRatio(num, denom):
    """``toRatio :: Integral a => a -> a -> Ratio a``

    Conversion to Ratio.

    """
    frac = fractions.Fraction(num, denom)
    return R(frac.numerator, frac.denominator)


instance(Real, int).where(
    toRational = lambda x: toRatio(x, 1)
)

try:
    instance(Real, long).where(
        toRational = lambda x: toRatio(x, 1)
    )
except NameError:
    pass

instance(Real, float).where(
    toRational = lambda x: toRatio(round(x), 1)  # obviously incorrect
)

instance(Integral, int).where(
    quotRem = lambda x, y: (x / y, x % y),
    toInteger = int,
    quot = lambda a, b: int(a)/int(b),
    rem = int.__mod__,
    div = lambda a, b: int(a)/int(b),
    mod = int.__mod__,
    divMod = divmod
)

try:
    instance(Integral, long).where(
        quotRem = lambda x, y: (x / y, x % y),
        toInteger = int,
        quot = long.__div__,
        rem = long.__mod__,
        div = long.__div__,
        mod = long.__mod__,
        divMod = divmod
    )
except NameError:
    pass


class RealFrac(Real, Fractional):
    """Extracting components of fractions.

    Dependencies:

    - `Real`:class:
    - `Fractional`:class:

    Attributes:

    - ``properFraction``
    - ``truncate``
    - ``round``
    - ``ceiling``
    - ``floor``

    Minimal complete definition:

    - ``properFraction``
    - ``truncate``
    - ``round``
    - ``ceiling``
    - ``floor``

    """
    @classmethod
    def make_instance(typeclass, cls, properFraction, truncate, round, ceiling,
            floor):
        attrs = {"properFraction": properFraction, "truncate": truncate,
                "round": round, "ceiling": ceiling, "floor": floor}
        build_instance(RealFrac, cls, attrs)


@sig(H[(RealFrac, "a"), (Integral, "b")]/ "a" >> ("b", "a"))
def properFraction(x):
    """``properFraction :: RealFrac a, Integral b => a -> (b, a)``

    The function properFraction takes a real fractional number x and returns a
    pair (n,f) such that x = n+f, and:

        n is an integral number with the same sign as x; and
        f is a fraction with the same type and sign as x, and with absolute
        value less than 1.

    """
    return RealFrac[x].properFraction(x)


@sig(H[(RealFrac, "a"), (Integral, "b")]/ "a" >> "b")
def truncate(x):
    """``truncate :: RealFrac a, Integral b => a -> b``

    truncate(x) returns the integer nearest x between zero and x

    """
    return RealFrac[x].truncate(x)


@sig(H[(RealFrac, "a"), (Integral, "b")]/ "a" >> "b")
def round(x):
    """``round :: RealFrac a, Integral b => a -> b``

    round(x) returns the nearest integer to x; the even integer if x is
    equidistant between two integers

    """
    return RealFrac[x].round(x)


@sig(H[(RealFrac, "a"), (Integral, "b")]/ "a" >> "b")
def ceiling(x):
    """``ceiling :: RealFrac a, Integral b => a -> b``

    ceiling(x) returns the least integer not less than x

    """
    return RealFrac[x].ceiling(x)


@sig(H[(RealFrac, "a"), (Integral, "b")]/ "a" >> "b")
def floor(x):
    """``floor :: RealFrac a, Integral b => a -> b``

    floor(x) returns the greatest integer not greater than x

    """
    return RealFrac[x].floor(x)


instance(RealFrac, float).where(
    properFraction = lambda x: (int(math.floor(x)), x - math.floor(x)),
    truncate = lambda x: int(math.floor(x) if x > 0 else math.floor(x+1)),
    round = lambda x: int(round(x, 0)),
    ceiling = math.ceil,
    floor = math.floor
)


class RealFloat(Floating, RealFrac):
    """Efficient, machine-independent access to the components of a
    floating-point number.

    Dependencies:

    - `Floating`:class:
    - `RealFrac`:class:

    Attributes:

    - ``floatRange``
    - ``isNan``
    - ``isInfinite``
    - ``isNegativeZero``
    - ``atan2``

    Minimal complete definition:

    - ``floatRange``
    - ``isNan``
    - ``isInfinite``
    - ``isNegativeZero``
    - ``atan2``

    """
    @classmethod
    def make_instance(typeclass, cls, floatRange, isNan, isInfinite,
            isNegativeZero, atan2):
        attrs = {"floatRange": floatRange, "isNan": isNan,
                "isInfinite": isInfinite, "isNegativeZero": isNegativeZero,
                "atan2": atan2}
        build_instance(RealFloat, cls, attrs)


@sig(H[(RealFloat, "a")]/ "a" >> bool)
def isNaN(x):
    """``isNaN :: RealFloat a => a -> bool``

    True if the argument is an IEEE "not-a-number" (NaN) value

    """
    return RealFloat[x].isNan(x)


@sig(H[(RealFloat, "a")]/ "a" >> bool)
def isInfinite(x):
    """``isInfinite :: RealFloat a => a -> bool``

    True if the argument is an IEEE infinity or negative infinity

    """
    return RealFloat[x].isInfinite(x)


@sig(H[(RealFloat, "a")]/ "a" >> bool)
def isNegativeZero(x):
    """``isNegativeZero :: RealFloat a => a -> bool``

    True if the argument is an IEEE negative zero

    """
    return RealFloat[x].isNegativeZero(x)


@sig(H[(RealFloat, "a")]/ "a" >> "a" >> "a")
def atan2(y, x):
    """``atan2 :: RealFloat a => a -> a -> a``

    a version of arctangent taking two real floating-point arguments

    """
    return RealFloat[x].atan2(y, x)


instance(RealFloat, float).where(
    floatRange=(sys.float_info.min, sys.float_info.max),
    isNan=math.isnan,
    isInfinite=lambda x: x == float('inf') or x == -float('inf'),
    isNegativeZero=lambda x: math.copysign(1, x) == -1.0,
    atan2=math.atan2
)



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ..lang import Read
from ..lang import Show
from ..lang import sig
from ..lang import H
from ..lang import t
from ..lang import d
from ..lang import caseof
from ..lang import m
from ..lang import p
from ..lang import data
from ..lang import deriving
from ..lang import instance
from ..lang import L
from ..lang import typify

from .Eq import Eq
from .Ord import Ord
from .Functor import Functor
from ..Control.Applicative import Applicative
from ..Control.Monad import Monad


# data Either a b = Left b | Right a deriving(Read, Show, Eq, Ord)
Either, Left, Right =\
        data.Either("a", "b") == d.Left("a") | d.Right("b") & \
        deriving(Read, Show, Eq, Ord)


instance(Functor, Either).where(
    fmap = lambda f, v: ~(caseof(v)
                            | m(Left(m.e))  >> Left(p.e)
                            | m(Right(m.ra)) >> Right(f(p.ra)))
)

instance(Applicative, Either).where(
    pure = Right
)

instance(Monad, Either).where(
    bind = lambda v, f: ~(caseof(v)
                            | m(Left(m.e))  >> Left(p.e)
                            | m(Right(m.a)) >> f(p.a))
)


def in_either(fn):
    """Decorator for monadic error handling.  If the decorated function raises an
    exception, return the exception inside Left. Otherwise, take the result
    and wrap it in Right.

    """
    def closure_in_either(*args, **kwargs):
        try:
            return Right(fn(*args, **kwargs))
        except Exception as e:
            return Left(e)
    return typify(fn, hkt=lambda x: t(Either, "aa", x))(closure_in_either)


@sig(H/ (H/ "a" >> "c") >> (H/ "b" >> "c") >> t(Either, "a", "b") >> "c")
def either(fa, fb, e):
    """``either :: (a -> c) -> (b -> c) -> Either a b -> c``

    Case analysis for the Either type. If the value is Left(a), apply the first
    function to a; if it is Right(b), apply the second function to b.

    """
    return ~(caseof(e)
                | m(Left(m.a))  >> fa(p.a)
                | m(Right(m.b)) >> fb(p.b))


@sig(H/ [t(Either, "a", "b")] >> ["a"])
def lefts(xs):
    """``lefts :: [Either a b] -> [a]``

    Extracts from a List of Either all the Left elements. All the Left elements
    are extracted in order.

    """
    return L[(x[0] for x in xs if isLeft(x))]


@sig(H/ [t(Either, "a", "b")] >> ["b"])
def rights(xs):
    """``rights :: [Either a b] -> [b]``

    Extracts from a list of Either all the Right elements. All the Right
    elements are extracted in order.

    """
    return L[(x[0] for x in xs if isRight(x))]


@sig(H/ t(Either, "a", "b") >> bool)
def isLeft(x):
    """``isLeft :: Either a b -> bool``

    Return True if the given value is a Left-value, False otherwise.

    """
    return ~(caseof(x)
                | m(Right(m.x)) >> False
                | m(Left(m.x))  >> True)


@sig(H/ t(Either, "a", "b") >> bool)
def isRight(x):
    """``isRight :: Either a b -> bool``

    Return True if the given value is a Right-value, False otherwise.

    """
    return not isLeft(x)


@sig(H/ [t(Either, "a", "b")] >> (["a"], ["b"]))
def partitionEithers(xs):
    """``partitionEithers :: [Either a b] -> ([a], [b])``

    Partitions a List of Either into two lists. All the Left elements are
    extracted, in order, to the first component of the output. Similarly the
    Right elements are extracted to the second component of the output.

    """
    return (lefts(xs), rights(xs))



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import functools
import types
import string
import sys
from collections import namedtuple, Mapping

from xoutil.eight.meta import metaclass
from xoutil.values.simple import logic_collection_coerce, nil as NIL


from .hindley_milner import TypeVariable
from .hindley_milner import TypeOperator
from .hindley_milner import Var
from .hindley_milner import App
from .hindley_milner import Lam
from .hindley_milner import unify
from .hindley_milner import analyze
from .hindley_milner import Function
from .hindley_milner import Tuple
from .hindley_milner import ListType


if sys.version[0] == '2':
    lowercase = string.lowercase
else:
    lowercase = string.ascii_lowercase


if sys.version[0] == '2':
    __python_builtins__ = set((
        types.BooleanType, types.BufferType, types.BuiltinFunctionType,
        types.BuiltinMethodType, types.ClassType, types.CodeType,
        types.ComplexType, types.DictProxyType, types.DictType,
        types.DictionaryType, types.EllipsisType, types.FileType,
        types.FloatType, types.FrameType, types.FunctionType,
        types.GeneratorType, types.GetSetDescriptorType, types.InstanceType,
        types.IntType, types.LambdaType, types.ListType, types.LongType,
        types.MemberDescriptorType, types.MethodType, types.ModuleType,
        types.NoneType, types.NotImplementedType, types.ObjectType,
        types.SliceType, types.StringType, types.StringTypes,
        types.TracebackType, types.TupleType, types.TypeType,
        types.UnboundMethodType, types.UnicodeType, types.XRangeType, set,
        frozenset))

    __python_function_types__ = set((
        types.FunctionType, types.LambdaType, types.MethodType,
        types.UnboundMethodType, types.BuiltinFunctionType,
        types.BuiltinMethodType))

else:
    __python_builtins__ = set((
        bool, dict, type(Ellipsis), float, int, type(None), str, tuple,
        complex, list, set, frozenset, slice,
        type, types.BuiltinFunctionType, types.BuiltinMethodType,
        types.CodeType, types.DynamicClassAttribute, types.FrameType,
        types.FunctionType, types.GeneratorType, types.GetSetDescriptorType,
        types.LambdaType, types.MappingProxyType, types.MemberDescriptorType,
        types.MethodType, types.ModuleType, types.TracebackType))

    __python_function_types__ = set((
        types.FunctionType, types.LambdaType, types.MethodType,
        types.BuiltinFunctionType, types.BuiltinMethodType))


def is_collection(m):
    '''True if `m` is a collection.

    strings are not collections.

    '''
    return logic_collection_coerce(m) is not NIL or isinstance(m, Mapping)


def is_builtin(cls):
    """
    Test whether a class or type is a Python builtin.

    Args:
        cls: The class or type to examine

    Returns:
        True if a type is a Python builtin type, and False otherwise.
    """
    return cls in __python_builtins__


def nt_to_tuple(nt):
    """
    Convert an instance of namedtuple (or an instance of a subclass of
    namedtuple) to a tuple, even if the instance's __iter__
    method has been changed. Useful for writing derived instances of
    typeclasses.

    Args:
        nt: an instance of namedtuple

    Returns:
        A tuple containing each of the items in nt
    """
    return tuple((getattr(nt, f) for f in nt.__class__._fields))


class TypeMeta(type):
    """
    Metaclass for Typeclass type. Ensures that all typeclasses are instantiated
    with a dictionary to map instances to their member functions, and a list of
    dependencies.
    """
    def __init__(self, *args):
        super(TypeMeta, self).__init__(*args)
        self.__instances__ = {}
        self.__dependencies__ = self.mro()[1:-2]  # excl self, Typeclass, object

    def __getitem__(self, item):
        try:
            if isinstance(item, ADT):
                return self.__instances__[id(item.__type_constructor__)]
            elif isinstance(typeof(item), ListType):
                return self.__instances__[id(type(item))]
            elif isinstance(item, Hask):
                return self.__instances__[id(typeof(item))]
            return self.__instances__[id(type(item))]
        except KeyError:
            raise TypeError("No instance for {0}".format(item))


class Typeclass(metaclass(TypeMeta)):
    """
    Base class for Hask typeclasses.

    All subclasses should implement make_instance, which controls what happens
    when a new instance is added. This method should set up whatever
    attributes/functions belong to the typeclass, and then call build_instance.
    See typeclasses.py for examples.
    """
    @classmethod
    def make_instance(typeclass, type_, *args):
        raise NotImplementedError("Typeclasses must implement make_instance")

    @classmethod
    def derive_instance(typeclass, type_):
        raise NotImplementedError("Typeclasses must implement derive_instance")


def build_instance(typeclass, cls, attrs):
    """
    Add a new instance to a typeclass, i.e. modify the typeclass's instance
    dictionary to include the new instance.

    Args:
        typeclass: The typeclass for which we are adding an instance
        cls: The class or type to be added
        attrs: A dict of {str:function}, mapping function names to functions
               for the typeclass instance

    Returns: None

    Raises:
        TypeError, if cls is not a member of all superclasses of typeclass
    """
    # 1) check dependencies
    for dep in typeclass.__dependencies__:
        if id(cls) not in dep.__instances__:
            raise TypeError("Missing dependency: %s" % dep.__name__)

    # 2) add type and its instance method to typeclass's instance dictionary
    __methods__ = namedtuple("__%s__" % str(id(cls)), attrs.keys())(**attrs)
    typeclass.__instances__[id(cls)] = __methods__


def has_instance(cls, typeclass):
    """
    Test whether a class is a member of a particular typeclass.

    Args:
        cls: The class or type to test for membership
        typeclass: The typeclass to check. Must be a subclass of Typeclass.

    Returns:
        True if cls is a member of typeclass, and False otherwise.
    """
    if not issubclass(typeclass, Typeclass):
        return False
    return id(cls) in typeclass.__instances__


class Hask(object):
    """Base class for objects within hask.

    `ADTs <ADT>`:class:, `TypedFunc`:class:,
    `~hask.lang.lazylist.List`:class:, `Undefined`:class:, and other
    hask-related types are all subclasses of Hask.

    All subclasses must define ``__type__``, which returns a representation of
    the object in the internal type system language.

    """
    def __type__(self):
        raise TypeError()


class Undefined(Hask):
    """
    A class with no concrete type definition (so its type can unify with any
    other type). Used to create `undefined` and to enable psuedo-laziness in
    pattern matching.
    """
    def __type__(self):
        return TypeVariable()


class PyFunc(object):
    """
    Singleton object that represents (any of the) Python function types in the
    type system and in type signatures.
    """
    pass


def typeof(obj):
    """Returns the type of an object within the internal type system.

    :param obj: the object to inspect

    :returns: An object representing the type in the internal type system
        language (i.e., a `~hask.lang.hindley_milner.TypeOperator`:class: or
        `~hask.lang.hindley_milner.TypeVariable`:class:).

    """
    TypeVariable.next_var_name = 'a'

    if isinstance(obj, Hask):
        return obj.__type__()

    elif isinstance(obj, tuple):
        return Tuple([typeof(o) for o in obj])

    elif obj is None:
        return TypeOperator(None, [])

    elif type(obj) in __python_function_types__:
        return TypeOperator(PyFunc, [])

    return TypeOperator(type(obj), [])


class TypeSignature(object):
    """Internal representation of a type signature, consisting of a list of
    function type arguments and a list of (typeclass, type_variable) typeclass
    constraint pairs.

    """
    def __init__(self, args, constraints):
        self.args = args
        self.constraints = constraints


class TypeSignatureHKT(object):
    """Internal representation of a higher-kinded type within a type signature,
    consisting of the type constructor and its type parameter names.

    """
    def __init__(self, tcon, params):
        self.tcon = tcon
        self.params = params


class TypeSignatureError(Exception):
    pass


def build_sig_arg(arg, cons, var_dict):
    """
    Covert a single argument of a type signature into its internal type system
    representation.

    Args:
        arg: The argument (a string, a Python type, etc) to convert
        cons: a dictionary of typeclass constraints for the type signature
        var_dict: a dictionary of bound type variables

    Returns: A TypeVariable or TypeOperator representing the arg

    Raises: TypeSignatureError, if the argument cannot be converted
    """
    # string representing type variable
    if isinstance(arg, str) and all((l in lowercase for l in arg)):
        if arg not in var_dict:
            if arg in cons:
                var_dict[arg] = TypeVariable(constraints=cons[arg])
            else:
                var_dict[arg] = TypeVariable()
        return var_dict[arg]

    # subsignature, e.g. H/ (H/ int >> int) >> int >> int
    elif isinstance(arg, TypeSignature):
        return make_fn_type(build_sig(arg, var_dict))

    # HKT, e.g. t(Maybe "a") or t("m", "a", "b")
    elif isinstance(arg, TypeSignatureHKT):
        if type(arg.tcon) == str:
            hkt = build_sig_arg(arg.tcon, cons, var_dict)
        else:
            hkt = arg.tcon
        return TypeOperator(hkt, [build_sig_arg(a, cons, var_dict)
                                  for a in arg.params])

    # None (the unit type)
    elif arg is None:
        return TypeOperator(None, [])

    # Tuples: ("a", "b"), (int, ("a", float)), etc.
    elif isinstance(arg, tuple):
        f = lambda x: build_sig_arg(x, cons, var_dict)
        return Tuple([f(x) for x in arg])

    # Lists: ["a"], [int], etc.
    elif isinstance(arg, list) and len(arg) == 1:
        return ListType(build_sig_arg(arg[0], cons, var_dict))

    # any other type, builtin or user-defined
    elif isinstance(arg, type):
        return TypeOperator(arg, [])

    raise TypeSignatureError("Invalid item in type signature: %s" % arg)


def make_fn_type(params):
    """
    Turn a list of type parameters into the corresponding internal type system
    object that represents the type of a function over the parameters.

    Args:
        params: a list of type paramaters, e.g. from a type signature. These
                should be instances of TypeOperator or TypeVariable

    Returns:
        An instance of TypeOperator representing the function type
    """
    if len(params) == 2:
        last_input, return_type = params
        return Function(last_input, return_type)
    return Function(params[0], make_fn_type(params[1:]))


def build_sig(type_signature, var_dict=None):
    """
    Parse a TypeSignature object and convert it to the internal type system
    language.

    Args:
        type_signature: an instance of TypeSignature
        var_dict: a dictionary of already-bound type variables, or None

    Returns: A list of TypeVariable/TypeOperator objects, representing the
             function type corresponding to the type signature
    """
    args = type_signature.args
    cons = type_signature.constraints
    var_dict = {} if var_dict is None else var_dict
    return [build_sig_arg(i, cons, var_dict) for i in args]


class TypedFunc(Hask):
    """Partially applied, statically typed function wrapper.

    """
    def __init__(self, fn, fn_args, fn_type):
        self.__doc__ = fn.__doc__
        self.func = fn
        self.fn_args = fn_args
        self.fn_type = fn_type

    def __type__(self):
        return self.fn_type

    def __call__(self, *args, **kwargs):
        # the environment contains the type of the function and the types
        # of the arguments
        env = {id(self): self.fn_type}
        env.update({id(arg): typeof(arg) for arg in args})
        ap = Var(id(self))

        for arg in args:
            if isinstance(arg, Undefined):
                return arg
            ap = App(ap, Var(id(arg)))
        result_type = analyze(ap, env)

        if len(self.fn_args) - 1 == len(args):
            result = self.func(*args)
            unify(result_type, typeof(result))
            return result
        return TypedFunc(functools.partial(self.func, *args, **kwargs),
                         self.fn_args[len(args):], result_type)

    def __mod__(self, arg):
        """
        (%) :: (a -> b) -> a -> b

        % is the apply operator, equivalent to $ in Haskell
        """
        return self.__call__(arg)

    def __mul__(self, fn):
        """
        (*) :: (b -> c) -> (a -> b) -> (a -> c)

        * is the function compose operator, equivalent to . in Haskell
        """
        if not isinstance(fn, TypedFunc):
            return fn.__rmul__(self)

        env = {id(self): self.fn_type, id(fn): fn.fn_type}
        compose = Lam("arg", App(Var(id(self)), App(Var(id(fn)), Var("arg"))))
        newtype = analyze(compose, env)

        composed_fn = lambda x: self.func(fn.func(x))
        newargs = [fn.fn_args[0]] + self.fn_args[1:]

        return TypedFunc(composed_fn, fn_args=newargs, fn_type=newtype)


class ADT(Hask):
    """Base class for Hask algebraic data types."""
    pass


def make_type_const(name, typeargs):
    """
    Build a new type constructor given a name and the type parameters.

    Args:
        name: the name of the new type constructor to be created
        typeargs: the type parameters to the constructor

    Returns:
        A new class that acts as a type constructor
    """
    def raise_fn(err):
        raise err()

    default_attrs = {"__params__": tuple(typeargs), "__constructors__": ()}
    cls = type(name, (ADT,), default_attrs)

    cls.__type__ = lambda self: \
        TypeOperator(cls, [TypeVariable() for i in cls.__params__])

    # Unless typeclass instances are provided or derived, ADTs do not support
    # any of these attributes, and trying to use one is a TypeError
    cls.__iter__ = lambda self: raise_fn(TypeError)
    cls.__contains__ = lambda self, other: raise_fn(TypeError)
    cls.__add__ = lambda self, other: raise_fn(TypeError)
    cls.__radd__ = lambda self, other: raise_fn(TypeError)
    cls.__rmul__ = lambda self, other: raise_fn(TypeError)
    cls.__mul__ = lambda self, other: raise_fn(TypeError)
    cls.__lt__ = lambda self, other: raise_fn(TypeError)
    cls.__gt__ = lambda self, other: raise_fn(TypeError)
    cls.__le__ = lambda self, other: raise_fn(TypeError)
    cls.__ge__ = lambda self, other: raise_fn(TypeError)
    cls.__eq__ = lambda self, other: raise_fn(TypeError)
    cls.__ne__ = lambda self, other: raise_fn(TypeError)
    cls.count = lambda self, other: raise_fn(TypeError)
    cls.index = lambda self, other: raise_fn(TypeError)

    # Unless Show is instantiated/derived, use object's `repr` method
    cls.__repr__ = object.__repr__
    cls.__str__ = object.__str__

    return cls


def make_data_const(name, fields, type_constructor, slot_num):
    """
    Build a data constructor given the name, the list of field types, and the
    corresponding type constructor.

    The general approach is to create a subclass of the type constructor and a
    new class created with `namedtuple`, with some of the features from
    `namedtuple` such as equality and comparison operators stripped out.

    Args:
        name: the name of the data constructor (a string)
        fields: the list of fields that the data constructor will have (a list
                of strings and types)
        type_constructor: the type constructor for the data constructor's type
        slot_num: the position of the data constructor in the `data` statement
                  defining the new type (e.g., Nothing=0, Just=1)

    Returns:
        The new data constructor
    """
    # create the data constructor
    base = namedtuple(name, ["i%s" % i for i, _ in enumerate(fields)])
    cls = type(name, (type_constructor, base), {})
    cls.__type_constructor__ = type_constructor
    cls.__ADT_slot__ = slot_num

    if len(fields) == 0:
        # If the data constructor takes no arguments, create an instance of it
        # and return that instance rather than returning the class
        cls = cls()
    else:
        # Otherwise, modify __type__ so that it matches up fields from the data
        # constructor with type params from the type constructor
        def __type__(self):
            args = [
                typeof(self[fields.index(p)]) if p in fields else TypeVariable()
                for p in type_constructor.__params__
            ]
            return TypeOperator(type_constructor, args)
        cls.__type__ = __type__

    type_constructor.__constructors__ += (cls,)
    return cls


def build_ADT(typename, typeargs, data_constructors, to_derive):
    """
    Create a new algebraic data type (a type constructor and at least one data
    constructor).

    Args:
        typename: a string representing the name of the type constructor
        typeargs: strings representing the type parameters of the type
                  constructor (should be unique, lowercase strings)
        data_constructors: a list of (name, [field]) pairs representing
                           each of the data constructors for the new type.
        to_derive: a list of typeclasses (subclasses of Typeclass) that should
                   be derived for the new type

    Returns:
        The type constructor, followed by each of the data constructors (in the
        order they were defined)

    Example usage:
        build_ADT(typename="Maybe",
                  typeargs=["a"],
                  data_constructors=[("Nothing", []), ("Just", ["a"])],
                  to_derive=[Read, Show, Eq, Ord])
    """
    # 1) Create the new type constructor and data constructors
    newtype = make_type_const(typename, typeargs)
    dcons = [make_data_const(dc_name, dc_fields, newtype, i)
             for i, (dc_name, dc_fields) in enumerate(data_constructors)]

    # 2) Derive typeclass instances for the new type constructor
    for tclass in to_derive:
        tclass.derive_instance(newtype)

    # 3) Wrap data constructors in TypedFunc instances
    for i, (dc_name, dc_fields) in enumerate(data_constructors):
        if len(dc_fields) == 0:
            continue

        return_type = TypeSignatureHKT(newtype, typeargs)
        sig = TypeSignature(list(dc_fields) + [return_type], [])
        sig_args = build_sig(sig, {})
        dcons[i] = TypedFunc(dcons[i], sig_args, make_fn_type(sig_args))
    return tuple([newtype, ] + dcons)


class PatternMatchBind(object):
    """Represents a local variable bound by pattern matching."""
    def __init__(self, name):
        self.name = name


class PatternMatchListBind(object):
    """
    Represents the head (first element) and tail (remaining elements) of a
    pattern-matched list
    """
    def __init__(self, head, tail):
        self.head = head
        self.tail = tail


def pattern_match(value, pattern, env=None):
    """Pattern match a value and a pattern.

    :param value: the value to pattern-match on

    :param pattern: a pattern, consisting of literals and/or locally bound
                 variables

    :param env: a dictionary of local variables bound while matching

    :returns: (True, env) if the match is successful, and (False, env)
              otherwise

    :raises SyntaxError: if a variable name is used multiple times in the same
        pattern

    """
    env = {} if env is None else env
    if isinstance(pattern, PatternMatchBind):
        if pattern.name in env:
            raise SyntaxError("Conflicting definitions for %s" % pattern.name)
        env[pattern.name] = value
        return True, env

    elif isinstance(pattern, PatternMatchListBind):
        head, tail = list(value[:len(pattern.head)]), value[len(pattern.head):]
        matches, env = pattern_match(head, pattern.head, env)
        if matches:
            return pattern_match(tail, pattern.tail, env)
        return False, env

    elif type(value) == type(pattern):
        if isinstance(value, ADT):
            return pattern_match(nt_to_tuple(value), nt_to_tuple(pattern), env)

        elif is_collection(value):
            matches = []
            if len(value) != len(pattern):
                return False, env

            for v, p in zip(value, pattern):
                match_status, env = pattern_match(v, p, env)
                matches.append(match_status)
            return all(matches), env

        elif value == pattern:
            return True, env

    return False, env



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ..lang import Read
from ..lang import Show
from ..lang import L
from ..lang import H
from ..lang import sig
from ..lang import t
from ..lang import data
from ..lang import d
from ..lang import caseof
from ..lang import m
from ..lang import p
from ..lang import deriving
from ..lang import instance
from ..lang import typify

from .Eq import Eq
from .Ord import Ord
from .Functor import Functor
from ..Control.Applicative import Applicative
from ..Control.Monad import Monad


# data Maybe a = Nothing | Just a deriving(Show, Eq, Ord)
Maybe, Nothing, Just =\
    data.Maybe("a") == d.Nothing | d.Just("a") & deriving(Read, Show, Eq, Ord)

instance(Functor, Maybe).where(
    fmap = lambda f, x: ~(caseof(x)
                            | m(Just(m.a)) >> Just(f(p.a))
                            | m(Nothing)   >> Nothing)
)

instance(Applicative, Maybe).where(
    pure = Just
)

instance(Monad, Maybe).where(
    bind = lambda x, f: ~(caseof(x)
                            | m(Just(m.a)) >> f(p.a)
                            | m(Nothing)   >> Nothing)
)


def in_maybe(fn):
    """Decorator for monadic error handling.

    If the decorated function raises an exception, return Nothing. Otherwise,
    take the result and wrap it in a Just.

    """
    def closure_in_maybe(*args, **kwargs):
        try:
            return Just(fn(*args, **kwargs))
        except:  # noqa
            return Nothing
    return typify(fn, hkt=lambda x: t(Maybe, x))(closure_in_maybe)


@sig(H/ "b" >> (H/ "a" >> "b") >> t(Maybe, "a") >> "b")
def maybe(default, f, maybe_a):
    """``maybe :: b -> (a -> b) -> Maybe a -> b``

    The maybe function takes a default value, a function, and a Maybe value. If
    the Maybe value is Nothing, the function returns the default value.
    Otherwise, it applies the function to the value inside the Just and returns
    the result.

    """
    return default if maybe_a == Nothing else f(maybe_a[0])


@sig(H/ t(Maybe, "a") >> bool)
def isJust(a):
    '``isJust :: [Maybe a] -> bool``'
    return not isNothing(a)


@sig(H/ t(Maybe, "a")  >> bool)
def isNothing(a):
    '``isNothing :: [Maybe a] -> bool``'
    return ~(caseof(a)
                | m(Nothing)   >> True
                | m(Just(m.x)) >> False)


@sig(H/ t(Maybe, "a") >> "a")
def fromJust(x):
    '``fromJust :: [Maybe a] -> a``'
    if isJust(x):
        return x[0]
    raise ValueError("Cannot call fromJust on Nothing.")


@sig(H/ ["a"] >> t(Maybe, "a"))
def listToMaybe(a):
    '``listToMaybe :: [a] -> [Maybe a]``'
    return ~(caseof(a)
                | m(m.a ^ m.b) >> Just(p.a)
                | m(m.a)       >> Nothing)


@sig(H/ t(Maybe, "a") >> ["a"])
def maybeToList(a):
    """``maybeToList :: Maybe a -> [a]``

    The maybeToList function returns an empty list when given Nothing or a
    singleton list when not given Nothing.

    """
    return ~(caseof(a)
                | m(Nothing)   >> L[[]]
                | m(Just(m.x)) >> L[[p.x]])


@sig(H/ [t(Maybe, "a")] >> ["a"])
def catMaybes(a):
    """``catMaybes :: [Maybe a] -> [a]``

    The catMaybes function takes a list of Maybes and returns a list of all the
    Just values.

    """
    return L[(fromJust(item) for item in a if isJust(item))]


@sig(H/ (H/ "a" >> t(Maybe, "b")) >> ["a"] >> ["b"])
def mapMaybe(f, la):
    """``mapMaybe :: (a -> Maybe b) -> [a] -> [b]``

    The mapMaybe function is a version of map which can throw out elements. In
    particular, the functional argument returns something of type Maybe b. If
    this is Nothing, no element is added on to the result list. If it is Just
    b, then b is included in the result list.

    """
    return L[(fromJust(b) for b in (f(a) for a in la) if isJust(b))]



'''Implementation of Hindley-Milner type inference system for Python, based
on Robert Smallshire's implementation for OWL BASIC.

Robert's original version can be found here:
http://smallshire.org.uk/sufficientlysmall/2010/04/11/a-hindley-milner-type-inference-implementation-in-python/

Changes from Robert's version:

1) Simplified type language somewhat (Let and Letrec are merged, as are Var
   and Ident)

2) Type system expanded to handle polymorphic higher-kinded types (however,
   in its current state it does not do this correctly due to the way that
   typeclasses were bolted on; this will be fixed in future versions)

3) Interface tweaked a bit to work better with Python types, including
   pretty-printing of type names and some useful subclasses of TypeOperator

4) Type unification also unifies typeclass constraints

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

# Class definitions for the AST nodes which comprise the type language for
# which types will be inferred


class AST(object):
    '''A low-level AST node in a typed lambda calculus.'''
    pass


class Lam(AST):
    '''Lambda abstraction'''

    def __init__(self, v, body):
        self.v = v
        self.body = body

    def __str__(self):
        return "(\{v} -> {body})".format(v=self.v, body=self.body)


class Var(AST):
    '''Variable/Identifier'''

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return str(self.name)


class App(AST):
    '''Function application.

    An App node represents the application of a **single** argument.
    Functions over several arguments are curried.

    '''
    def __init__(self, fn, arg):
        self.fn = fn
        self.arg = arg

    def __str__(self):
        return "({fn} {arg})".format(fn=self.fn, arg=self.arg)


class Let(AST):
    '''Let binding (always recursive)'''

    def __init__(self, v, defn, body):
        self.v = v
        self.defn = defn
        self.body = body

    def __str__(self):
        exp = "(let {v} = {defn} in {body})"
        return exp.format(v=self.v, defn=self.defn, body=self.body)


def show_type(type_name):
    '''Pretty-print a Python type or internal type name.

    :param type_name: a Python type, or a string representing a type name

    :returns: a string representation of the type

    '''
    if isinstance(type_name, str):
        return type_name
    elif isinstance(type_name, type):
        return type_name.__name__
    return str(type_name)


class TypeVariable(object):
    '''A type variable standing for an arbitrary type.

    All type variables have a unique id, but names are only assigned lazily,
    when required.

    Note that this approach is *not* thread-safe.

    '''

    next_variable_id = 0
    next_var_name = 'a'

    def __init__(self, constraints=()):
        self.id = TypeVariable.next_variable_id
        TypeVariable.next_variable_id += 1
        self.instance = None
        self.__name = None
        self.constraints = constraints

    def __getName(self):
        '''
        Names are allocated to TypeVariables lazily, so that only TypeVariables
        converted to strings are given names.
        '''
        if self.__name is None:
            self.__name = TypeVariable.next_var_name
            TypeVariable.next_var_name = chr(ord(TypeVariable.next_var_name) + 1)
        return self.__name

    name = property(__getName)

    def __str__(self):
        if self.instance is not None:
            return str(self.instance)
        return self.name

    def __repr__(self):
        return "TypeVariable(id = {0})".format(self.id)


class TypeOperator(object):
    '''An n-ary type constructor which builds a new type from old'''

    def __init__(self, name, types):
        self.name = name
        self.types = types

    def __str__(self):
        num_types = len(self.types)
        if num_types == 0:
            return show_type(self.name)
        return "({0} {1})".format(
            show_type(self.name),
            ' '.join(show_type(t) for t in self.types)
        )


class Function(TypeOperator):
    '''A binary type constructor which builds function types'''

    def __init__(self, from_type, to_type):
        super(self.__class__, self).__init__("->", [from_type, to_type])

    def __str__(self):
        return "({1} {0} {2})".format(
            show_type(self.name),
            show_type(self.types[0]),
            show_type(self.types[1]),
        )


class Tuple(TypeOperator):
    '''N-ary constructor which builds tuple types'''

    def __init__(self, types):
        super(self.__class__, self).__init__(tuple, types)

    def __str__(self):
        return "({0})".format(", ".join(map(show_type, self.types)))


class ListType(TypeOperator):
    '''Unary constructor which builds list types'''

    def __init__(self, list_type):
        super(self.__class__, self).__init__("[]", [list_type])

    def __str__(self):
        return "[{0}]".format(show_type(self.types[0]))


def analyze(node, env, non_generic=None):
    '''Computes the type of the expression given by node.

    The type of the node is computed in the context of the supplied type
    environment, env.  Data types can be introduced into the language simply
    by having a predefined set of identifiers in the initial environment.
    This way there is no need to change the syntax or, more importantly, the
    type-checking program when extending the language.

    :param node: The root of the abstract syntax tree.

    :param env: The type environment is a mapping of expression identifier
                names to type assignments.  to type assignments.

    :param non_generic: A set of non-generic variables, or None

    :returns: The computed type of the expression.

    :raises TypeError: The type of the expression could not be inferred, for
         example if it is not possible to unify two types such as Integer and
         Bool or if the abstract syntax tree rooted at node could not be
         parsed

    '''
    if non_generic is None:
        non_generic = set()
    if isinstance(node, Var):
        return getType(node.name, env, non_generic)
    elif isinstance(node, App):
        fun_type = analyze(node.fn, env, non_generic)
        arg_type = analyze(node.arg, env, non_generic)
        result_type = TypeVariable()
        unify(Function(arg_type, result_type), fun_type)
        return result_type
    elif isinstance(node, Lam):
        arg_type = TypeVariable()
        new_env = env.copy()
        new_env[node.v] = arg_type
        new_non_generic = non_generic.copy()
        new_non_generic.add(arg_type)
        result_type = analyze(node.body, new_env, new_non_generic)
        return Function(arg_type, result_type)
    elif isinstance(node, Let):
        new_type = TypeVariable()
        new_env = env.copy()
        new_env[node.v] = new_type
        new_non_generic = non_generic.copy()
        new_non_generic.add(new_type)
        defn_type = analyze(node.defn, new_env, new_non_generic)
        unify(new_type, defn_type)
        return analyze(node.body, new_env, non_generic)
    else:
        assert False, "Unhandled syntax node {0}".format(node)


def getType(name, env, non_generic):
    '''Get the type of identifier name from the type environment env.

    :param name: The identifier name
    :param env: The type environment mapping from identifier names to types
    :param non_generic: A set of non-generic TypeVariables

    :raises ParseError: Raised if name is an undefined symbol in the type
            environment.
    '''
    if name in env:
        return fresh(env[name], non_generic)
    raise TypeError("Undefined symbol {0}".format(name))


def fresh(t, non_generic):
    '''Makes a copy of a type expression.

    The type t is copied.  The the generic variables are duplicated and the
    non_generic variables are shared.

    :param t: A type to be copied.
    :param non_generic: A set of non-generic TypeVariables

    '''
    mappings = {}  # A mapping of TypeVariables to TypeVariables

    def freshrec(tp):
        p = prune(tp)
        if isinstance(p, TypeVariable):
            if isGeneric(p, non_generic):
                if p not in mappings:
                    mappings[p] = TypeVariable()
                return mappings[p]
            else:
                return p
        elif isinstance(p, TypeOperator):
            return TypeOperator(p.name, [freshrec(x) for x in p.types])

    return freshrec(t)


def unify_var(v1, t2):
    '''Unify the type variable `v1` and the type `t2`, i.e. makes their types the
    same and unifies typeclass constraints.  Note: Must be called with v1 and
    t2 pre-pruned

    :param v1: The type variable to be made equivalent
    :param t2: The second type to be be equivalent

    Instead of returning, modify `v1` and `t2` to add the unification as a
    constraint.

    :raises TypeError: Raised if the types cannot be unified.

    '''
    if v1 != t2:
        if isinstance(t2, TypeVariable):
            # unify typeclass constraints
            union = tuple(set(v1.constraints + t2.constraints))
            v1.constraints = union
            t2.constraints = union

        if occursInType(v1, t2):
            raise TypeError("recursive unification")
        v1.instance = t2


def unify(t1, t2):
    '''Unify the two types t1 and t2.  Makes the types t1 and t2 the same.

    Note that the current method of unifying higher-kinded types does not
    properly handle kind, i.e. it will happily unify `f a` and `g b c`.
    This is due to the way that typeclasses are implemented, and will be fixed
    in future versions.

    :param t1: The first type to be made equivalent
    :param t2: The second type to be be equivalent

    Modify `t1` and `t2` in-place, by modifying their constraints for the
    unification.

    :raises TypeError: Raised if the types cannot be unified.

    '''
    a = prune(t1)
    b = prune(t2)
    if isinstance(a, TypeVariable):
        unify_var(a, b)
    elif isinstance(a, TypeOperator) and isinstance(b, TypeVariable):
        unify_var(b, a)
    elif isinstance(a, TypeOperator) and isinstance(b, TypeOperator):
        # Unify polymorphic higher-kinded type
        if isinstance(a.name, TypeVariable) and len(a.types) > 0:
            a.name = b.name
            a.types = b.types
            unify(a, b)
        elif isinstance(b.name, TypeVariable):
            unify(b, a)

        # Unify concrete higher-kinded type
        elif (a.name != b.name or len(a.types) != len(b.types)):
            raise TypeError("Type mismatch: {0} != {1}".format(str(a), str(b)))
        for p, q in zip(a.types, b.types):
            unify(p, q)
    else:
        raise TypeError("Not unified")


def prune(t):
    '''Returns the currently defining instance of t.

    As a side effect, collapses the list of type instances.  The function
    prune is used whenever a type expression has to be inspected: it will
    always return a type expression which is either an uninstantiated type
    variable or a type operator; i.e. it will skip instantiated variables, and
    will actually prune them from expressions to remove long chains of
    instantiated variables.

    :param t: The type to be pruned

    :returns: An uninstantiated TypeVariable or a TypeOperator

    '''
    if isinstance(t, TypeVariable):
        if t.instance is not None:
            t.instance = prune(t.instance)
            return t.instance
    return t


def isGeneric(v, non_generic):
    '''
    Checks whether a given variable occurs in a list of non-generic variables

    Note that a variables in such a list may be instantiated to a type term,
    in which case the variables contained in the type term are considered
    non-generic.

    .. note:: Must be called with `v` pre-pruned

    :param v: The TypeVariable to be tested for genericity
    :param non_generic: A set of non-generic TypeVariables

    :returns: True if v is a generic variable, otherwise False

    '''
    return not occursIn(v, non_generic)


def occursInType(v, type2):
    '''Checks whether a type variable occurs in a type expression.

    .. note:: Must be called with v pre-pruned

    :param v:  The TypeVariable to be tested for
    :param type2: The type in which to search

    :returns: True if v occurs in type2, otherwise False

    '''
    pruned_type2 = prune(type2)
    if pruned_type2 == v:
        return True
    elif isinstance(pruned_type2, TypeOperator):
        return occursIn(v, pruned_type2.types)
    return False


def occursIn(t, types):
    '''Checks whether a types variable occurs in any other types.

    :param v:  The TypeVariable to be tested for
    :param types: The sequence of types in which to search

    :returns: True if t occurs in any of types, otherwise False

    '''
    return any(occursInType(t, t2) for t2 in types)



# flake8: noqa
from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from .typeclasses import Show
from .typeclasses import show
from .typeclasses import Read
from .typeclasses import Eq
from .typeclasses import Ord
from .typeclasses import Bounded
from .lazylist import Enum
from .lazylist import succ
from .lazylist import pred
from .lazylist import fromEnum
from .lazylist import enumFrom
from .lazylist import enumFromTo
from .lazylist import enumFromThen
from .lazylist import enumFromThenTo

from .type_system import typeof
from .type_system import is_builtin
from .type_system import has_instance
from .type_system import nt_to_tuple
from .type_system import build_instance
from .type_system import Typeclass
from .type_system import Hask
from .type_system import TypedFunc
from .type_system import TypeSignatureError

from .syntax import undefined
from .syntax import caseof
from .syntax import m
from .syntax import p
from .syntax import IncompletePatternError
from .syntax import data
from .syntax import d
from .syntax import deriving
from .syntax import H
from .syntax import sig
from .syntax import t
from .syntax import func
from .syntax import typify
from .syntax import NoGuardMatchException
from .syntax import guard
from .syntax import c
from .syntax import otherwise
from .syntax import instance
from .syntax import __
from .syntax import _t
from .syntax import _q
from .syntax import _i

from .lazylist import List
from .lazylist import L



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ..lang import sig
from ..lang import H
from ..lang import t


# Current implementation is just a wrapper around Python's Fraction. This is
# not a long-term solution (not extensible beyond builtin types) but it will do
# for now.

from .Num import Integral
from .Num import RealFrac
from .Num import Ratio
from .Num import R  # noqa
from .Num import Rational
from .Num import toRatio
from .Num import toRational  # noqa


@sig(H[(Integral, "a")]/ t(Ratio, "a") >> "a")
def numerator(ratio):
    """``numerator :: Integral a => Ratio a -> a``

    Extract the numerator of the ratio in reduced form: the numerator and
    denominator have no common factor and the denominator is positive.

    """
    return toRatio(ratio[0], ratio[1])[0]


@sig(H[(Integral, "a")]/ t(Ratio, "a") >> "a")
def denominator(ratio):
    """``denominator :: Integral a => Ratio a -> a``

    Extract the denominator of the ratio in reduced form: the numerator and
    denominator have no common factor and the denominator is positive.

    """
    return toRatio(ratio[0], ratio[1])[1]


@sig(H[(RealFrac, "a")]/ "a" >> "a" >> Rational)
def approxRational(x, epsilon):
    """``approxRational :: RealFrac a => a -> a -> Rational``

    approxRational, applied to two real fractional numbers x and epsilon,
    returns the simplest rational number within epsilon of x. A rational
    number y is said to be simpler than another y' if ``abs(numerator(y)) <=
    abs(numerator(y_))`` and ``denominator(y) <= denominator(y_)``.

    Any real interval contains a unique simplest rational; in particular, note
    that 0/1 is the simplest rational of all.

    """
    raise NotImplementedError()



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import collections
import itertools
import sys

from .hindley_milner import TypeVariable
from .hindley_milner import ListType
from .hindley_milner import unify

from .type_system import typeof
from .type_system import Typeclass
from .type_system import Hask
from .type_system import build_instance

from .typeclasses import Show
from .typeclasses import show
from .typeclasses import Eq
from .typeclasses import Ord

from .syntax import Syntax
from .syntax import instance
from .syntax import sig
from .syntax import H

try:
    from __builtin__ import cmp
except ImportError:
    def cmp(a, b):
        if a == b:
            return 0
        elif a < b:
            return -1
        else:
            return 1


class Enum(Typeclass):
    """
    Class Enum defines operations on sequentially ordered types.

    The enumFrom... methods are used in translation of arithmetic sequences.

    Instances of Enum may be derived for any enumeration type (types whose
    constructors have no fields). The nullary constructors are assumed to be
    numbered left-to-right by fromEnum from 0 through n-1.

    Attributes:

    - ``toEnum``
    - ``fromEnum``
    - ``succ``
    - ``pred``
    - ``enumFrom``
    - ``enumFromThen``
    - ``enumFrom``
    - ``enumFromThenTo``
    - ``EnumFromTo``

    Minimal complete definition:

    - ``toEnum``
    - ``fromEnum``

    """
    @classmethod
    def make_instance(typeclass, cls, toEnum, fromEnum):
        def succ(a):
            return toEnum(fromEnum(a) + 1)

        def pred(a):
            return toEnum(fromEnum(a) - 1)

        def enumFromThen(start, second):
            pointer = fromEnum(start)
            step = fromEnum(second) - pointer
            while True:
                yield toEnum(pointer)
                pointer += step

        def enumFrom(start):
            return enumFromThen(start, succ(start))

        def enumFromThenTo(start, second, end):
            if start == end:
                yield start
                return

            elif (second >= start > end) or (second <= start < end):
                return

            pointer, stop = fromEnum(start), fromEnum(end)
            step = fromEnum(second) - pointer
            while (start < end and pointer <= stop) or \
                  (start > end and pointer >= stop):
                yield toEnum(pointer)
                pointer += step

        def enumFromTo(start, end):
            second = succ(start) if start < end else pred(start)
            return enumFromThenTo(start, second, end)

        attrs = {"toEnum": toEnum, "fromEnum": fromEnum, "succ": succ, "pred":
                 pred, "enumFromThen": enumFromThen, "enumFrom": enumFrom,
                 "enumFromThenTo": enumFromThenTo, "enumFromTo": enumFromTo}
        build_instance(Enum, cls, attrs)


@sig(H/ "a" >> int)
def fromEnum(a):
    """``fromEnum :: a -> int``

    Convert to an int.

    """
    return Enum[a].toEnum(a)


@sig(H/ "a" >> "a")
def succ(a):
    """``succ :: a -> a``

    the successor of a value. For numeric types, succ adds 1.

    """
    return Enum[a].succ(a)


@sig(H/ "a" >> "a")
def pred(a):
    """
    pred :: a -> a

    the predecessor of a value. For numeric types, pred subtracts 1.
    """
    return Enum[a].pred(a)


@sig(H/ "a" >> "a" >> ["a"])
def enumFromThen(start, second):
    """``enumFromThen :: a -> a -> [a]``

    Used in translation of [n, n_, ...]

    """
    return L[Enum[start].enumFromThen(start, second)]


@sig(H/ "a" >> ["a"])
def enumFrom(start):
    """``enumFrom :: a -> [a]``

    Used in translation of L[n, ...]

    """
    return L[Enum[start].enumFrom(start)]


@sig(H/ "a" >> "a" >> "a" >> ["a"])
def enumFromThenTo(start, second, end):
    """``enumFromThenTo :: a -> a -> a -> [a]``

    Used in translation of L[n, n_, ..., m]

    """
    return L[Enum[start].enumFromThenTo(start, second, end)]


@sig(H/ "a" >> "a" >> ["a"])
def enumFromTo(start, end):
    """``enumFromTo :: a -> a -> [a]``

    Used in translation of L[n, ..., m]

    """
    return L[Enum[start].enumFromTo(start, end)]


instance(Enum, int).where(fromEnum=int, toEnum=int)
instance(Enum, bool).where(fromEnum=int, toEnum=bool)
instance(Enum, str).where(fromEnum=ord, toEnum=chr)

if sys.version[0] == '2':
    long = long   # noqa
    instance(Enum, long).where(fromEnum=int, toEnum=long)


class List(collections.Sequence, Hask):
    """Statically typed lazy sequence datatype.

    See `L`:obj: for more information.

    """
    def __init__(self, head=None, tail=None):
        self.__head = []
        self.__tail = itertools.chain([])
        self.__is_evaluated = True

        if head is not None and len(head) > 0:
            fst = head[0]
            for fst, other in zip(itertools.repeat(fst), head):
                unify(typeof(fst), typeof(other))
            self.__head.extend(head)
        if tail is not None:
            self.__tail = itertools.chain(self.__tail, tail)
            self.__is_evaluated = False

    def __type__(self):
        if self.__is_evaluated:
            if len(self.__head) == 0:
                return ListType(TypeVariable())
            return ListType(typeof(self[0]))

        elif len(self.__head) == 0:
            self.__next()
            return self.__type__()

        return ListType(typeof(self[0]))

    def __next(self):
        """
        Evaluate the next element of the tail, and add it to the head.
        """
        if self.__is_evaluated:
            raise StopIteration
        else:
            try:
                next_iter = next(self.__tail)
                if len(self.__head) > 0:
                    unify(typeof(self[0]), typeof(next_iter))
                self.__head.append(next_iter)
            except StopIteration:
                self.__is_evaluated = True

    def __evaluate(self):
        """
        Evaluate the entire List.
        """
        while not self.__is_evaluated:
            self.__next()

    def __rxor__(self, item):
        """
        ^ is the cons operator (equivalent to : in Haskell)
        """
        unify(self.__type__(), ListType(typeof(item)))
        if self.__is_evaluated:
            return List(head=[item] + self.__head)
        return List(head=[item] + self.__head, tail=self.__tail)

    def __add__(self, other):
        """
        (+) :: [a] -> [a] -> [a]

        + is the list concatenation operator, equivalent to ++ in Haskell and +
        for Python lists
        """
        unify(self.__type__(), typeof(other))
        if self.__is_evaluated and other.__is_evaluated:
            return List(head=self.__head + other.__head)
        elif self.__is_evaluated and not other.__is_evaluated:
            return List(head=self.__head + other.__head,
                        tail=other.__tail)
        return List(head=self.__head,
                    tail=itertools.chain(self.__tail, iter(other)))

    def __str__(self):
        if len(self.__head) == 0 and self.__is_evaluated:
            return "L[[]]"

        elif len(self.__head) == 1 and self.__is_evaluated:
            return "L[[%s]]" % show(self.__head[0])

        body = ", ".join((show(s) for s in self.__head))
        return "L[%s]" % body if self.__is_evaluated else "L[%s ...]" % body

    def __cmp__(self, other):
        if self.__is_evaluated and other.__is_evaluated:
            return cmp(self.__head, other.__head)

        elif len(self.__head) >= len(other.__head):
            # check the evaluated heads
            heads = zip(self.__head[:len(other.__head)], other.__head)
            heads_comp = ((cmp(h1, h2) for h1, h2 in heads))
            for comp in heads_comp:
                if comp != 0:
                    return comp

            # evaluate the shorter-headed list until it is the same size
            while len(self.__head) > len(other.__head):
                if other.__is_evaluated:
                    return 1
                other.__next()
                comp = cmp(self.__head[len(other.__head)-1], other.__head[-1])
                if comp != 0:
                    return comp

            # evaluate the tails, checking each time
            while not self.__is_evaluated or not other.__is_evaluated:
                if not self.__is_evaluated:
                    self.__next()
                if not other.__is_evaluated:
                    other.__next()

                len_comp = cmp(len(self.__head), len(other.__head))
                if len_comp != 0:
                    return len_comp

                if len(self.__head) > 0:
                    value_comp = cmp(self.__head[-1], other.__head[-1])
                    if value_comp != 0:
                        return value_comp

        elif len(other.__head) > len(self.__head):
            return -other.__cmp__(self)

        return 0

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        return self.__cmp__(other) == -1

    def __gt__(self, other):
        return self.__cmp__(other) == 1

    def __le__(self, other):
        comp = self.__cmp__(other)
        return comp in (-1, 0)

    def __ge__(self, other):
        comp = self.__cmp__(other)
        return comp in (1, 0)

    def __len__(self):
        self.__evaluate()
        return len(self.__head)

    def __iter__(self):
        for item in self.__head:
            yield item

        for item in self.__tail:
            self.__head.append(item)
            yield item

    def count(self, x):
        unify(self.__type__(), ListType(typeof(x)))
        self.__evaluate()
        return self.__head.count(x)

    def index(self, x):
        unify(self.__type__(), ListType(typeof(x)))
        self.__evaluate()
        return self.__head.index(x)

    def __contains__(self, x):
        unify(self.__type__(), ListType(typeof(x)))
        for item in iter(self):
            if item is x:
                return True
        return False

    def __getitem__(self, ix):
        is_slice = isinstance(ix, slice)
        if is_slice:
            i = ix.start if ix.stop is None else ix.stop
        else:
            i = ix
        # make sure that the list is evaluated enough to do the indexing, but
        # not any more than necessary
        # if index is negative, evaluate the entire list
        if i is None:
            # In Python 3, `None >= 0` is a TypeError, but in Python 2 returns
            # False.  So let's go negative in any case...
            i = -1
        if i >= 0:
            while (i+1) > len(self.__head):
                try:
                    self.__next()
                except StopIteration:
                    break
        else:
            self.__evaluate()

        if is_slice:
            if ix.stop is None and not self.__is_evaluated:
                return List(head=self.__head[ix], tail=self.__tail)
            return List(head=self.__head[ix])
        return self.__head[i]


# Basic typeclass instances for list
instance(Show, List).where(
    show = List.__str__
)

instance(Eq, List).where(
    eq = List.__eq__
)

instance(Ord, List).where(
    lt = List.__lt__,
    gt = List.__gt__,
    le = List.__le__,
    ge = List.__ge__
)


class __list_comprehension__(Syntax):
    """
    L is the syntactic construct for Haskell-style list comprehensions and lazy
    list creation. To create a new List, just wrap an interable in L[ ].

    List comprehensions can be used with any instance of Enum, including the
    built-in types int, long, float, and char.
    There are four basic list comprehension patterns:

    >>> L[1, ...]
    # list from 1 to infinity, counting by ones

    >>> L[1, 3, ...]
    # list from 1 to infinity, counting by twos

    >>> L[1, ..., 20]
    # list from 1 to 20 (inclusive), counting by ones

    >>> L[1, 5, ..., 20]
    # list from 1 to 20 (inclusive), counting by fours

    """
    def __getitem__(self, lst):
        if isinstance(lst, tuple) and len(lst) < 5 and \
                any((Ellipsis is x for x in lst)):
            # L[x, ...]
            if len(lst) == 2 and lst[1] is Ellipsis:
                return enumFrom(lst[0])

            # L[x, y, ...]
            elif len(lst) == 3 and lst[2] is Ellipsis:
                return enumFromThen(lst[0], lst[1])

            # L[x, ..., y]
            elif len(lst) == 3 and lst[1] is Ellipsis:
                return enumFromTo(lst[0], lst[2])

            # L[x, y, ..., z]
            elif len(lst) == 4 and lst[2] is Ellipsis:
                return enumFromThenTo(lst[0], lst[1], lst[3])

            raise SyntaxError("Invalid list comprehension: %s" % str(lst))

        elif hasattr(lst, "next") or hasattr(lst, "__next__"):
            return List(tail=lst)

        return List(head=lst)


L = __list_comprehension__("Invalid input to list constructor")



'''Typed wrappers for builtin Python functions.

This makes it easier to chain lots of things together in function composition
without having to manually add type signatures to Python builtins.

Each function is a `~hask.lang.type_system.TypedFunc`:class: replacement of
the corresponding Python builtin with the right signature.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from ..lang import H

try:
    from __builtin__ import cmp as pycmp
except ImportError:
    def pycmp(a, b):
        if a == b:
            return 0
        elif a < b:
            return -1
        else:
            return 1

callable = callable ** (H/ "a" >> bool)
cmp = pycmp ** (H/ "a" >> "a" >> int)
delattr = delattr ** (H/ "a" >> str >> None)
divmod = divmod ** (H/ "a" >> "b" >> ("c", "c"))
getattr = getattr ** (H/ "a" >> str >> "b")
hasattr = hasattr ** (H/ "a" >> str >> bool)
hash = hash ** (H/ "a" >> int)
hex = hex ** (H/ int >> str)
isinstance = isinstance ** (H/ "a" >> "b" >> bool)
issubclass = issubclass ** (H/ "a" >> "b" >> bool)
len = len ** (H/ "a" >> int)
oct = oct ** (H/ int >> str)
repr = repr ** (H/ "a" >> str)
setattr = setattr ** (H/ "a" >> str >> "b" >> None)
sorted = sorted ** (H/ "a" >> list)
try:
    from __builtin__ import unichr as pyunichr
except ImportError:
    pyunichr = chr
    unicode = str

unichr = pyunichr ** (H/ int >> unicode)



from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import operator
import sys

from .type_system import Typeclass
from .type_system import is_builtin
from .type_system import nt_to_tuple
from .type_system import build_instance

from .syntax import instance
from .syntax import sig
from .syntax import H


class Show(Typeclass):
    """Conversion of values to readable strings.

    Attributes:

    - ``__str__``
    - ``show``

    Minimal complete definition:

    - ``show``

    """
    @classmethod
    def make_instance(typeclass, cls, show):
        __show__ = show ** (H/ "a" >> str)
        show = lambda self: __show__(self)

        build_instance(Show, cls, {"show": show})
        if not is_builtin(cls):
            cls.__repr__ = show
            cls.__str__ = show

    @classmethod
    def derive_instance(typeclass, cls):
        def show(self):
            if len(self.__class__._fields) == 0:
                return self.__class__.__name__

            nt_tup = nt_to_tuple(self)
            if len(nt_tup) == 1:
                tuple_str = "(%s)" % Show[nt_tup[0]].show(nt_tup[0])
            else:
                tuple_str = Show[nt_tup].show(nt_tup)

            return "{0}{1}".format(self.__class__.__name__, tuple_str)
        Show.make_instance(cls, show=show)


@sig(H/ "a" >> str)
def show(obj):
    """``show :: a -> str``

    Convert a value to a readable string.

    """
    return Show[obj].show(obj)


class Eq(Typeclass):
    """The Eq class defines equality (==) and inequality (!=).

    Attributes:

    - ``__eq__``
    - ``__ne__``

    Minimal complete definition:

    - ``eq``

    """
    @classmethod
    def make_instance(typeclass, cls, eq, ne=None):
        def default_ne(self, other):
            return not eq(self, other)

        __eq__ = eq ** (H/ "a" >> "a" >> bool)
        __ne__ = (default_ne if ne is None else ne) ** (H/ "a" >> "a" >> bool)
        eq = lambda self, other: __eq__(self, other)
        ne = lambda self, other: __ne__(self, other)

        build_instance(Eq, cls, {"eq": eq, "ne": ne})
        if not is_builtin(cls):
            cls.__eq__ = eq
            cls.__ne__ = ne

    @classmethod
    def derive_instance(typeclass, cls):
        def __eq__(self, other):
            return self.__class__ == other.__class__ and \
                nt_to_tuple(self) == nt_to_tuple(other)

        def __ne__(self, other):
            return self.__class__ != other.__class__ or  \
                nt_to_tuple(self) != nt_to_tuple(other)

        Eq.make_instance(cls, eq=__eq__, ne=__ne__)


class Ord(Eq):
    """The Ord class is used for totally ordered datatypes.

    Instances of Ord can be derived for any user-defined datatype whose
    constituent types are in Ord. The declared order of the constructors in the
    data declaration determines the ordering in derived Ord instances. The
    Ordering datatype allows a single comparison to determine the precise
    ordering of two objects.

    Dependencies:

    - `Eq`:class:

    Attributes:

    - ``__lt__``
    - ``__le__``
    - ``__gt__``
    - ``__ge__``

    Minimal complete definition:

    - ``lt``

    """
    @classmethod
    def make_instance(typeclass, cls, lt, le=None, gt=None, ge=None):
        def_le = lambda s, o: s.__lt__(o) or s.__eq__(o)
        def_gt = lambda s, o: not s.__lt__(o) and not s.__eq__(o)
        def_ge = lambda s, o: not s.__lt__(o) or s.__eq__(o)

        __lt__ = lt ** (H/ "a" >> "a" >> bool)
        __le__ = (def_le if le is None else le) ** (H/ "a" >> "a" >> bool)
        __gt__ = (def_gt if gt is None else gt) ** (H/ "a" >> "a" >> bool)
        __ge__ = (def_ge if ge is None else ge) ** (H/ "a" >> "a" >> bool)

        lt = lambda self, other: __lt__(self, other)
        le = lambda self, other: __le__(self, other)
        gt = lambda self, other: __gt__(self, other)
        ge = lambda self, other: __ge__(self, other)

        attrs = {"lt": lt, "le": le, "gt": gt, "ge": ge}
        build_instance(Ord, cls, attrs)
        if not is_builtin(cls):
            cls.__lt__ = lt
            cls.__le__ = le
            cls.__gt__ = gt
            cls.__ge__ = ge

    @classmethod
    def derive_instance(typeclass, cls):
        def zip_cmp(self, other, fn):
            """
            Compare the data constructor and all of the fields of two ADTs.
            """
            if self.__ADT_slot__ == other.__ADT_slot__:
                if len(nt_to_tuple(self)) == 0:
                    return fn((), ())
                return fn(nt_to_tuple(self), nt_to_tuple(other))
            return fn(self.__ADT_slot__, other.__ADT_slot__)

        lt = lambda s, o: zip_cmp(s, o, operator.lt)
        le = lambda s, o: zip_cmp(s, o, operator.le)
        gt = lambda s, o: zip_cmp(s, o, operator.gt)
        ge = lambda s, o: zip_cmp(s, o, operator.ge)

        Ord.make_instance(cls, lt=lt, le=le, gt=gt, ge=ge)


class Bounded(Typeclass):
    """The Bounded class is used to name the upper and lower limits of a
    type. Ord is not a superclass of Bounded since types that are not totally
    ordered may also have upper and lower bounds.

    The Bounded class may be derived for any enumeration type; minBound is the
    first constructor listed in the data declaration and maxBound is the last.
    Bounded may also be derived for single-constructor datatypes whose
    constituent types are in Bounded.

    Attributes:

    - ``minBound``
    - ``maxBound``

    Minimal complete definition:

    - ``minBound``
    - ``maxBound``

    """
    @classmethod
    def make_instance(typeclass, cls, minBound, maxBound):
        attrs = {"minBound": minBound, "maxBound": maxBound}
        build_instance(Bounded, cls, attrs)

    @classmethod
    def derive_instance(typeclass, cls):
        # All data constructors must be enums
        for data_con in cls.__constructors__:
            if not isinstance(data_con, cls):
                raise TypeError("Cannot derive Bounded; %s is not an enum" %
                                 data_con.__name__)

        maxBound = lambda s: cls.__constructors__[0]
        minBound = lambda s: cls.__constructors__[-1]
        Bounded.make_instance(cls, minBound=minBound, maxBound=maxBound)


class Read(Typeclass):
    """Parsing of Strings, producing values.

    Attributes:

    - ``read``

    Minimal complete definition:

    - ``read``

    """
    @classmethod
    def make_instance(typeclass, cls, read):
        build_instance(Read, cls, {"read": read})

    @classmethod
    def derive_instance(typeclass, cls):
        Read.make_instance(cls, read=eval)


instance(Show, str).where(show=str.__repr__)
instance(Show, int).where(show=int.__str__)
instance(Show, float).where(show=tuple.__str__)
instance(Show, complex).where(show=complex.__str__)
instance(Show, bool).where(show=bool.__str__)
instance(Show, list).where(show=list.__str__)
instance(Show, tuple).where(show=tuple.__str__)
instance(Show, set).where(show=set.__str__)
instance(Show, dict).where(show=dict.__str__)
instance(Show, frozenset).where(show=frozenset.__str__)
instance(Show, slice).where(show=slice.__str__)

instance(Eq, str).where(eq=str.__eq__, ne=str.__ne__)
instance(Eq, int).where(eq=int.__eq__, ne=int.__ne__)
instance(Eq, float).where(eq=float.__eq__, ne=float.__ne__)
instance(Eq, complex).where(eq=complex.__eq__, ne=complex.__ne__)
instance(Eq, bool).where(eq=bool.__eq__, ne=bool.__ne__)
instance(Eq, list).where(eq=list.__eq__, ne=list.__ne__)
instance(Eq, tuple).where(eq=tuple.__eq__, ne=tuple.__ne__)
instance(Eq, set).where(eq=set.__eq__, ne=set.__ne__)
instance(Eq, dict).where(eq=dict.__eq__, ne=dict.__ne__)
instance(Eq, frozenset).where(eq=frozenset.__eq__, ne=frozenset.__ne__)
instance(Eq, slice).where(eq=slice.__eq__, ne=slice.__ne__)
instance(Eq, type).where(eq=type.__eq__, ne=type.__ne__)

instance(Ord, str).where(lt=str.__lt__, le=str.__le__,
                         gt=str.__gt__, ge=str.__ge__)
instance(Ord, int).where(lt=int.__lt__, le=int.__le__,
                         gt=int.__gt__, ge=int.__ge__)
instance(Ord, float).where(lt=float.__lt__, le=float.__le__,
                           gt=float.__gt__, ge=float.__ge__)
instance(Ord, complex).where(lt=complex.__lt__, le=complex.__le__,
                             gt=complex.__gt__, ge=complex.__ge__)
instance(Ord, bool).where(lt=bool.__lt__, le=bool.__le__,
                          gt=bool.__gt__, ge=bool.__ge__)
instance(Ord, list).where(lt=list.__lt__, le=list.__le__,
                          gt=list.__gt__, ge=list.__ge__)
instance(Ord, tuple).where(lt=tuple.__lt__, le=tuple.__le__,
                           gt=tuple.__gt__, ge=tuple.__ge__)
instance(Ord, set).where(lt=set.__lt__, le=set.__le__,
                          gt=set.__gt__, ge=set.__ge__)
instance(Ord, dict).where(lt=dict.__lt__, le=dict.__le__,
                          gt=dict.__gt__, ge=dict.__ge__)
instance(Ord, frozenset).where(lt=frozenset.__lt__, le=frozenset.__le__,
                               gt=frozenset.__gt__, ge=frozenset.__ge__)

if sys.version[0] == '2':
    instance(Show, long).where(show=long.__str__)
    instance(Show, unicode).where(show=unicode.__str__)

    instance(Eq, long).where(eq=long.__eq__, ne=long.__ne__)
    instance(Eq, unicode).where(eq=unicode.__eq__, ne=unicode.__ne__)

    instance(Ord, long).where(lt=long.__lt__, le=long.__le__,
                              gt=long.__gt__, ge=long.__ge__)
    instance(Ord, unicode).where(lt=unicode.__lt__, le=unicode.__le__,
                                gt=unicode.__gt__, ge=unicode.__ge__)
